//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  dkmsinstallmgr.cpp: DKMS installation manager
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "dkmsinstallmgr.h"
#include "pipereader.h"
#include <iostream>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#ifdef HAVE_RPM
#include <rpmlib.h>
#include <rpmdb.h>
#include <rpmts.h>
#include <rpmio.h>
#endif

using namespace gmodcfg;
using namespace std;

//--------------------------------------------------------------------------//
//  C'tor
//--------------------------------------------------------------------------//
DkmsInstallManager::DkmsInstallManager( int verbosity )
    : InstallManager( verbosity ) {}

#ifdef HAVE_RPM

//--------------------------------------------------------------------------//
//  Helper function that extracts infos from an RPM header
//--------------------------------------------------------------------------//
static void
getRPMPackageInfos( Header hdr, char*& name, char*& version, char*& vendor,
                 char*& summary )
{
    int_32 type, count;
    headerGetEntry( hdr, RPMTAG_NAME, &type, (void **) &name, &count );
    headerGetEntry( hdr, RPMTAG_VERSION, &type, (void **) &version, &count );
    headerGetEntry( hdr, RPMTAG_VENDOR, &type, (void **) &vendor, &count );
    headerGetEntry( hdr, RPMTAG_SUMMARY, &type, (void **) &summary, &count );
}

//--------------------------------------------------------------------------//
//  RPM install progression callback
//--------------------------------------------------------------------------//
static void*
progressCallback( const void*, const rpmCallbackType type,
                  const unsigned long long amount, const unsigned long long total,
                  const void *, void* data )
{
    InstallManager* mgr = static_cast<InstallManager*>( data );

    switch( type )
    {
        case RPMCALLBACK_INST_OPEN_FILE:
        case RPMCALLBACK_INST_CLOSE_FILE:
            break;

        case RPMCALLBACK_INST_START:
        case RPMCALLBACK_INST_PROGRESS:
            mgr->UpdateProgressBar( (gfloat)amount / (gfloat)total );

        default: break;
    }
    return NULL;
}

//--------------------------------------------------------------------------//
//  Install an RPM-based DKMS package
//--------------------------------------------------------------------------//
static bool
installRPM( rpmts ts, Header hdr, const string& pkgFile, int verbosity,
            DkmsInstallManager* mgr )
{
    // FIXME: use RPM's C API

    string cmd = "rpm -Uvh " + pkgFile;
    int rc = system( cmd.c_str() );

    if ( rc != 0 )
    {
        mgr->WriteStatusMessage( _("Problems occured, installation failed") );
        return false;
    }

    return true;
}

#endif

//--------------------------------------------------------------------------//
//  Install a DKMS tarball
//--------------------------------------------------------------------------//
static bool
installTarball( const string& pkgFile, int verbosity,
                DkmsInstallManager* mgr )
{
    // TODO: extract module and version infos and display them in window
 
    // FIXME: don't hardcode path to dkms
    string cmd = "/sbin/dkms ldtarball --archive=" + pkgFile;
    if ( verbosity >= 2 )
        cout << cmd << endl;

    int rc = system( cmd.c_str() );

    if ( verbosity >= 3 )
        cout << "rc = " << rc << endl;

    if ( rc )
    {
        mgr->WriteStatusMessage( _("An error occurred, installation failed") );
        return false;
    }

    return true;
}

//--------------------------------------------------------------------------//
//  Install a package, RPM or tarball
//--------------------------------------------------------------------------//
bool
DkmsInstallManager::installPackage( const string& pkgFile )
{
    UpdateProgressBar( 0 );

    if ( Verbosity )
        cout << "DKMS InstallManager: install package " << pkgFile << endl;

    bool installOK = false;

#ifdef HAVE_RPM
    rpmReadConfigFiles( NULL, NULL );
    rpmSetVerbosity( RPMMESS_VERBOSE );

    // Check if it's an RPM package by attempting to read the RPM header
    // If not an RPM, try to install it as a DKMS tarball
    rpmts ts = rpmtsCreate();
    FD_t fd = Fopen( pkgFile.c_str(), "r" );
    Header hdr;

    if ( rpmReadPackageFile( ts, fd, pkgFile.c_str(), &hdr ) == 0 )
    {
        // Show package infos in install mgr window
        char *name, *version, *vendor, *summary;
        getRPMPackageInfos( hdr, name, version, vendor, summary );
        showPackageInfos( name, version, vendor, summary );

        installOK = installRPM( ts, hdr, pkgFile, Verbosity, this );
    }
    else
#endif
        installOK = installTarball( pkgFile, Verbosity, this );

#ifdef HAVE_RPM
    Fclose( fd );
    headerFree( hdr );
    rpmtsFree( ts );
#endif
    UpdateProgressBar( 1.0 );

    return installOK;
}

//--------------------------------------------------------------------------//
//  Uninstall a package
//--------------------------------------------------------------------------//
bool
DkmsInstallManager::uninstallPackage( const string& pkgName )
{
    // TODO
    if ( Verbosity )
        cout << "DkmsInstallManager: uninstallPackage " << pkgName << endl;
    return false;
}

//--------------------------------------------------------------------------//
//  Check if a package is installed
//--------------------------------------------------------------------------//
bool
DkmsInstallManager::isPackageInstalled( const string& pkgName ) const
{
    // TODO
    if ( Verbosity )
        cout << "DkmsInstallManager: isPackageInstalled " << pkgName << endl;
    return false;
}

