//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  gmodapp.cpp: main application code
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "gmodapp.h"
#include "interface.h"
#include "support.h"

#include <iostream>
#include <cstdio>
#include <stdexcept>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/param.h>

#include <bonobo.h>

using namespace gmodcfg;
using namespace std;

enum { MODULE_NAME_COLUMN, SHORT_DESC_COLUMN, NUM_COLUMNS };

//--------------------------------------------------------------------------//
//  C'tor
//--------------------------------------------------------------------------//
Application::Application( const string& modPath, const string& modConfFile,
                          int v )
            : TopWidget(0), TreeView(0), StatusBar(0), SearchEntry(0),
              ModulesConfFile(modConfFile),
              ModulesPath(modPath), Verbosity(v) {}

//--------------------------------------------------------------------------//
//  Helper function: excludes . and .. from our directory search (scandir)
//--------------------------------------------------------------------------//
static int
filterFiles( const struct direct* entry )
{
    return ! ( (strcmp(entry->d_name, ".") == 0)
            || (strcmp(entry->d_name, "..") == 0) );
}

//--------------------------------------------------------------------------//
//  At a given directory level in the modules hierarchy, create an entry in
//  the tree_store and recursively call ourself on the files/dirs it contains.
//--------------------------------------------------------------------------//
static void
populateTreestore( GtkTreeStore* treestore, set<string>& modulesNames,
                   GtkTreeIter* parentIter, const string& pathName,
                   bool topLevel = false )
{
    struct direct** namelist;
    int count = scandir( pathName.c_str(), &namelist, filterFiles, 0 );
    const char* itemName = basename( pathName.c_str() );
 
    GtkTreeIter iter;

    // Start with only the subdirs below pathName when at the top
    if ( !topLevel )
    {
        gtk_tree_store_append( treestore, &iter, parentIter );

        // If no files found: this is a leaf (module)
        if ( count <= 0 )
        {
            // TODO: display module's short description in 2nd column
 
            // Strip the suffix off of the module name
            string str( itemName );
            int dot = str.find_last_of(".");
            if ( dot != string::npos )
                str.erase( dot, str.length()-dot );

	    // Second strip, as the first strip only removes .gz
	    // suffix from gzip'ed Mandrake Linux kernel modules
            dot = str.find_last_of(".");
            if ( dot != string::npos )
                str.erase( dot, str.length()-dot );

            gtk_tree_store_set( treestore, &iter, MODULE_NAME_COLUMN,
                                str.c_str(), SHORT_DESC_COLUMN, "", -1 );
            modulesNames.insert( str );
            return;
        }

        // That's a directory
        gtk_tree_store_set( treestore, &iter, MODULE_NAME_COLUMN, itemName,
                            SHORT_DESC_COLUMN, "", -1 );
    }

    for( int i=0; i<count; ++i )
    {
        const char* childName = namelist[i]->d_name;
        string childPath = pathName + "/" + childName;

        // go further down, in case this is a dir
        populateTreestore( treestore, modulesNames, ( topLevel ? 0 : &iter ),
                           childPath );
    }
}

//--------------------------------------------------------------------------//
//  Create 2 columns for our treeview
//--------------------------------------------------------------------------//
static void
addColumns( GtkTreeView* treeview, GtkTreeStore* treestore )
{
    GtkTreeModel* model = GTK_TREE_MODEL(treestore);
    gtk_tree_view_set_model( treeview, model );

    // 1st column: module names
    GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
    g_object_set( G_OBJECT(renderer), "xalign", 0.0, NULL );
    gint col_offset = gtk_tree_view_insert_column_with_attributes(
                            GTK_TREE_VIEW(treeview), -1, "Module",
                            renderer, "text", MODULE_NAME_COLUMN, NULL );
    GtkTreeViewColumn* column = gtk_tree_view_get_column(
                                GTK_TREE_VIEW(treeview), col_offset - 1 );
    gtk_tree_view_column_set_clickable( GTK_TREE_VIEW_COLUMN(column), TRUE );

    // 2nd column: short description
    renderer = gtk_cell_renderer_text_new();
    g_object_set( G_OBJECT(renderer), "xalign", 0.0, NULL );
    g_object_set_data( G_OBJECT(renderer), "column", (gint*)SHORT_DESC_COLUMN );
    col_offset = gtk_tree_view_insert_column_with_attributes(
                            GTK_TREE_VIEW(treeview), -1, "Short Description",
                            renderer, "text", SHORT_DESC_COLUMN, NULL );
    column = gtk_tree_view_get_column( GTK_TREE_VIEW(treeview),
                                        col_offset - 1 );
    gtk_tree_view_column_set_clickable( GTK_TREE_VIEW_COLUMN(column), TRUE );
}

//--------------------------------------------------------------------------//
//  Treeview's "changed" callback
//--------------------------------------------------------------------------//
static void
selectionChangedCallback( GtkTreeSelection* select, gpointer data )
{
    GtkTreeView* treeView = gtk_tree_selection_get_tree_view( select );
    GtkTreeModel* model = gtk_tree_view_get_model( treeView );
    Application* app = static_cast<Application*>(data);
    GtkTreeIter iter;
    if ( gtk_tree_selection_get_selected( select, &model, &iter ) )
    {
        char* itemName;
        gtk_tree_model_get( model, &iter, MODULE_NAME_COLUMN, &itemName, -1 );

        // read module's information and update the GUI
        app->BuildModuleGui( itemName );
    }
}

//--------------------------------------------------------------------------//
//  Application's "delete_event" callback
//--------------------------------------------------------------------------//
static void
deleteEventCallback( GtkWidget*, GdkEvent*, gpointer data )
{
    Application* app = static_cast<Application*>(data);
    app->Quit();
}

//--------------------------------------------------------------------------//
//  Helpers to get/set int and string properties on a Bonobo control
//--------------------------------------------------------------------------//

// Declared but not defined in the general case. See below for specializations.
template <typename T>
static void
setProp( Bonobo_PropertyBag, const char*, T val );

// Specialization for string properties
template <>
static void
setProp( Bonobo_PropertyBag pb, const char* propName, const char* val )
{
    bonobo_pbclient_set_string( pb, propName, val, CORBA_OBJECT_NIL );
}

// Specialization for int properties
template <>
static void
setProp( Bonobo_PropertyBag pb, const char* propName, int val )
{
    bonobo_pbclient_set_long( pb, propName, val, CORBA_OBJECT_NIL );
}

// The following is generic code to set the value of a property in a Bonobo
// property bag. The setProp() function hides the fact that Bonobo has
// different functions to set properties of different types. It's called
// sendCommand() since it's how the control gets... controlled :-)
template <typename T>
static void
sendCommand( GtkWidget* w, const char* propName, T propVal )
{
    CORBA_Environment ev;
    CORBA_exception_init( &ev );

    BonoboControlFrame* cf = bonobo_widget_get_control_frame(BONOBO_WIDGET(w));
    Bonobo_PropertyBag pb =
                bonobo_control_frame_get_control_property_bag ( cf, &ev );

    if ( BONOBO_EX(&ev) )
        g_warning( "Error getting property bag from control" );
    else
        setProp( pb, propName, propVal );

    bonobo_object_release_unref( pb, NULL );
    CORBA_exception_free( &ev );
}

// Declared but not defined in the general case. See below for specializations.
template <typename T>
static void
getProp( Bonobo_PropertyBag, const char*, T& );

template <>
static void
getProp( Bonobo_PropertyBag pb, const char* propName, gchar*& val )
{
    val = bonobo_pbclient_get_string( pb, propName, CORBA_OBJECT_NIL );
}

template <>
static void
getProp( Bonobo_PropertyBag pb, const char* propName, bool& val )
{
    val = bonobo_pbclient_get_boolean( pb, propName, CORBA_OBJECT_NIL );
}

template <typename T>
static void
getData( GtkWidget* w, const char* propName, T& val )
{
    CORBA_Environment ev;
    CORBA_exception_init( &ev );

    BonoboControlFrame* cf = bonobo_widget_get_control_frame(BONOBO_WIDGET(w));
    Bonobo_PropertyBag pb =
                bonobo_control_frame_get_control_property_bag ( cf, &ev );

    if ( BONOBO_EX(&ev) )
        g_warning( "Error getting property bag from control" );
    else
        getProp( pb, propName, val );

    bonobo_object_release_unref( pb, NULL );
    CORBA_exception_free( &ev );
}

//--------------------------------------------------------------------------//
//  Callback called when the control posts a new message
//--------------------------------------------------------------------------//
// KLUDGE: default BONOBO_ARG_GET_STRING breaks with g++
#define MY_BONOBO_ARG_GET_STRING(a)    (*((CORBA_char **)(a->_value)))

static void
on_new_control_message( BonoboListener*, const char*, const BonoboArg* arg,
                        CORBA_Environment*, gpointer data )
{
    GnomeAppBar* statusBar = static_cast<GnomeAppBar*>(data);
    gnome_appbar_set_status( statusBar, MY_BONOBO_ARG_GET_STRING(arg) );
}

//--------------------------------------------------------------------------//
//  Load and set-up the gmodconfig control
//--------------------------------------------------------------------------//
bool
Application::loadControl( GtkBox* vbox )
{
    bind_textdomain_codeset( GETTEXT_PACKAGE, "UTF-8" );

    const char* oafId = "OAFIID:GNOME_Linux_ModuleManager";
    ModuleManagerWidget = bonobo_widget_new_control( oafId, CORBA_OBJECT_NIL );
    if ( !ModuleManagerWidget )
        return false;

    gtk_widget_show( ModuleManagerWidget );
    gtk_box_pack_start( vbox, ModuleManagerWidget, TRUE, TRUE, 5 );

    // Register a listener on the control's event source
    bonobo_event_source_client_add_listener(
               bonobo_widget_get_objref( BONOBO_WIDGET(ModuleManagerWidget) ),
               on_new_control_message, "message", NULL, StatusBar );

    // Configure the control
    sendCommand( ModuleManagerWidget, "setModulesConfFile",
                       ModulesConfFile.c_str() );
    sendCommand( ModuleManagerWidget, "setVerbosity", Verbosity );

    return true;
}

//--------------------------------------------------------------------------//
//  Builds the static GUI and enters the GTK event loop
//--------------------------------------------------------------------------//
void
Application::Start()
{
    TopWidget = create_app();
    gtk_widget_show( TopWidget );
    StatusBar = GNOME_APPBAR( lookup_widget( TopWidget, "appbar" ) );
    SearchEntry = GNOME_ENTRY( lookup_widget( TopWidget, "search_entry" ) );
    GtkBox* vbox = GTK_BOX( lookup_widget( TopWidget, "rightvbox" ) );

    if ( !loadControl( vbox ) )
    {
        gnome_appbar_set_status( StatusBar, "Can't create control!" );
        return;
    }

    // delete_event happens when the user closes the window
    gtk_signal_connect( GTK_OBJECT( TopWidget ), "delete_event",
                        GTK_SIGNAL_FUNC( deleteEventCallback ),
                        (gpointer)this );

    // Populate the tree view
    TreeView = GTK_TREE_VIEW( lookup_widget( TopWidget, "treeview" ) );
    GtkTreeStore* treestore = gtk_tree_store_new( NUM_COLUMNS,
                                                G_TYPE_STRING, G_TYPE_STRING );
    populateTreestore( treestore, ModulesNames, NULL, ModulesPath, true );
    addColumns( TreeView, treestore );

    // Install treeview's "changed" callback
    GtkTreeSelection* select = gtk_tree_view_get_selection( TreeView );
    g_signal_connect( G_OBJECT( select ), "changed",
                      G_CALLBACK( selectionChangedCallback ), (gpointer)this );

    gtk_main();
}

//--------------------------------------------------------------------------//
//  Pop up a dialog if some changes are not yet saved
//--------------------------------------------------------------------------//
void
Application::Quit()
{
    bool changes = false;
    getData( ModuleManagerWidget, "hasUnsavedChanges", changes );

    if ( changes )
    {
        GtkWidget* quit = create_quitdialog();
        gtk_window_set_transient_for( GTK_WINDOW(quit), GTK_WINDOW(TopWidget) );
        gtk_window_present( GTK_WINDOW(quit) );
    }
    else
        gtk_main_quit();
}

//--------------------------------------------------------------------------//
//  Update module name and usage at top of window
//--------------------------------------------------------------------------//
static void
updateNameAndUsage( GtkWidget* topWidget, const string& name,
                    const string& usage )
{
    GtkLabel* modnameLabel = GTK_LABEL( lookup_widget(topWidget, "modname") );
    gtk_label_set_text( modnameLabel, name.c_str() );

    GtkLabel* usageLabel = GTK_LABEL( lookup_widget(topWidget, "usage") );
    gtk_label_set_text( usageLabel, usage.c_str() );
}

//--------------------------------------------------------------------------//
//  Dynamically fill in the control's tabs
//  Returns true for success.
//--------------------------------------------------------------------------//
bool
Application::BuildModuleGui( const char* modName )
{
    if ( !modName )
        return false;

    // If it's not in our internal modules list it's a directory
    if ( ModulesNames.count( modName ) == 0 )
        return false;

    // Tell the control to display that module
    sendCommand( ModuleManagerWidget, "showModule", modName );

    // Display the short description of that module
    gchar* usage = NULL;
    getData( ModuleManagerWidget, "getModuleUsage", usage );
    updateNameAndUsage( TopWidget, modName, usage );

    return true;
}

//--------------------------------------------------------------------------//
//  Add module name to search history (search_entry widget)
//--------------------------------------------------------------------------//
void
Application::AddToSearchHistory( const char* modName )
{
    // FIXME: duplication of entries is fixed in GNOME 2.2
    std::pair< SetString_t::iterator, bool > p = History.insert( modName );
    if ( p.second )
    {
        if ( Verbosity >= 3 )
            cout << "Adding " << modName << " to search history" << endl;

        gnome_entry_prepend_history( SearchEntry, FALSE, g_strdup(modName) );
    }
}

//--------------------------------------------------------------------------//
//  Save options to /etc/modules.conf
//--------------------------------------------------------------------------//
void
Application::SaveAllOptions()
{
    sendCommand( ModuleManagerWidget, "saveOptions", "" );
}

//--------------------------------------------------------------------------//
//  Bring up the install/uninstall manager dialogs
//--------------------------------------------------------------------------//
void
Application::InstallPackage()
{
    sendCommand( ModuleManagerWidget, "installPackage", "" );
}

void
Application::UnInstallPackage()
{
    sendCommand( ModuleManagerWidget, "unInstallPackage", "" );
}

//--------------------------------------------------------------------------//
//  Show the about box
//--------------------------------------------------------------------------//
void
Application::ShowAboutBox() const
{
    GtkWidget* about = create_about();
    gtk_window_set_transient_for( GTK_WINDOW(about), GTK_WINDOW(TopWidget) );
    gtk_window_present( GTK_WINDOW(about) );
}

