//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  installmgr.cpp: module installation manager
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "installmgr.h"
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>

#include <iostream>

using namespace gmodcfg;
using namespace std;

//--------------------------------------------------------------------------//
//  C'tor
//--------------------------------------------------------------------------//
InstallManager::InstallManager( int v )
        : InstallDialog(NULL), UnInstallDialog(NULL), InstallPathEntry(NULL),
          UnInstallEntry(NULL), PkgNameLabel(NULL), PkgVersionLabel(NULL),
          PkgVendorLabel(NULL), PkgSummaryLabel(NULL),
          InstallStatusLabel(NULL), UnInstallStatusLabel(NULL),
          ProgressBar(NULL), Verbosity(v)
{
}

//--------------------------------------------------------------------------//
//  Helper function to update the progress and status widgets
//--------------------------------------------------------------------------//
void
InstallManager::UpdateProgressBar( gfloat percent ) const
{
    gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR(ProgressBar), percent );
}

void
InstallManager::WriteStatusMessage( const char* msg ) const
{
    gtk_label_set_text( InstallStatusLabel, msg );
}

//--------------------------------------------------------------------------//
//  Progress callback: updates the progress bar
//--------------------------------------------------------------------------//
static gint
progressCallback( GnomeVFSXferProgressInfo* info, gpointer data )
{
    GnomeVFSFileSize fileSize = info->file_size;
    GnomeVFSFileSize copied = info->bytes_copied;
    InstallManager* mgr = static_cast<InstallManager*>(data);

    if ( fileSize )
    {
        gfloat progress = (gfloat)copied / (gfloat)fileSize;
        mgr->UpdateProgressBar( progress );

        while ( gtk_events_pending() )
            gtk_main_iteration(); // used to refresh GUI
    }
    return TRUE;
}

//--------------------------------------------------------------------------//
//  Helper function that creates empty labels in the table
//--------------------------------------------------------------------------//
static GtkLabel*
makeInfoLabel( GtkWidget* table, int row, int col )
{
    GtkWidget* label = gtk_label_new("");
    gtk_table_attach( GTK_TABLE(table), label, col, 2, row, row+1,
                     (GtkAttachOptions) (GTK_EXPAND | GTK_SHRINK | GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0 );
    GTK_WIDGET_SET_FLAGS( label, GTK_CAN_FOCUS );
    gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_LEFT );
    gtk_label_set_selectable( GTK_LABEL(label), TRUE );
    gtk_misc_set_alignment( GTK_MISC(label), 0, 0.5 );
    return GTK_LABEL(label);
}

//--------------------------------------------------------------------------//
//  Create and display the installer dialog window
//--------------------------------------------------------------------------//
void
InstallManager::makeInstallDialogBox()
{
    InstallDialog = GTK_DIALOG(
        gtk_dialog_new_with_buttons( _("Module Installation"), NULL,
                                    GTK_DIALOG_MODAL, 
                                    GTK_STOCK_HELP, GTK_RESPONSE_HELP,
                                    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                    GTK_STOCK_OK, GTK_RESPONSE_OK, NULL ) );

    GtkWidget* vbox = InstallDialog->vbox;
    GtkWidget* table = gtk_table_new( 10, 2, FALSE );
    gtk_box_pack_start( GTK_BOX(vbox), table, TRUE, TRUE, 0 );
    gtk_container_set_border_width( GTK_CONTAINER(table), 10 );
    gtk_table_set_row_spacings( GTK_TABLE(table), 2 );
    gtk_table_set_col_spacings( GTK_TABLE(table), 5 );

    int row = 0;

    // Label above the file entry widget
    GtkWidget* label1 = gtk_label_new( _("Select a package to install and click OK:") );
    gtk_table_attach( GTK_TABLE(table), label1, 0, 2, row, row+1,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0 );
    gtk_label_set_justify( GTK_LABEL(label1), GTK_JUSTIFY_LEFT );
    gtk_misc_set_alignment( GTK_MISC(label1), 0.1, 0.5 );
    gtk_misc_set_padding( GTK_MISC(label1), 10, 0 );
    ++row;

    // The URI entry widget
    GtkWidget* align = gtk_alignment_new( 0.5, 0.5, 0.8, 1 );
    gtk_table_attach( GTK_TABLE(table), align, 0, 2, row, row+1,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0 );
    gtk_container_set_border_width( GTK_CONTAINER(align), 10 );

    GtkWidget* uriEntry = gnome_file_entry_new( NULL, _("Select a package to install") );
    gtk_container_add( GTK_CONTAINER(align), uriEntry );
    gnome_file_entry_set_modal( GNOME_FILE_ENTRY(uriEntry), TRUE );

    InstallPathEntry = GTK_ENTRY( gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY(uriEntry) ) );
    gtk_entry_set_width_chars( InstallPathEntry, 60 );
    ++row;

    GtkWidget* sep1 = gtk_hseparator_new();
    gtk_table_attach( GTK_TABLE(table), sep1, 0, 2, row, row+1,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0 );
    ++row;

    // Info labels showing details about the package being installed
    const gchar* labelNames[] = { _("Name:"), _("Version:"), _("Vendor:"),
                                  _("Summary:") };
    for( int i=0; i<4; ++i )
    {
        GtkWidget* label = gtk_label_new( labelNames[i] );
        gtk_table_attach( GTK_TABLE(table), label, 0, 1, row+i, row+i+1,
                         (GtkAttachOptions) (GTK_FILL),
                         (GtkAttachOptions) (0), 0, 0 );
        gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_LEFT );
        gtk_misc_set_alignment( GTK_MISC(label), 0.95, 0.5 );
    }

    PkgNameLabel = makeInfoLabel( table, row++, 1 );
    PkgVersionLabel = makeInfoLabel( table, row++, 1 );
    PkgVendorLabel = makeInfoLabel( table, row++, 1 );
    PkgSummaryLabel = makeInfoLabel( table, row++, 1 );

    GtkWidget* sep2 = gtk_hseparator_new();
    gtk_table_attach( GTK_TABLE(table), sep2, 0, 2, row, row+1,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (GTK_FILL), 0, 0 );
    ++row;

    // Status label and progress bar
    InstallStatusLabel = makeInfoLabel( table, row++, 0 );
    ProgressBar = gtk_progress_bar_new();
    gtk_table_attach( GTK_TABLE(table), ProgressBar, 0, 2, row, row+1,
                     (GtkAttachOptions) (GTK_EXPAND | GTK_SHRINK | GTK_FILL),
                     (GtkAttachOptions) (0), 0, 10 );

    gtk_widget_show_all( GTK_WIDGET(InstallDialog) );
    gtk_window_present( GTK_WINDOW(InstallDialog) );

    // Wait for user's response, might get more than one button click
    bool keepGoing = true;
    while ( keepGoing )
    {
        switch( gtk_dialog_run( InstallDialog ) )
        {
            case GTK_RESPONSE_OK:
                DoInstall();
                break;

            case GTK_RESPONSE_HELP:
                // TODO
                break;

            default: // cancel or delete event
                keepGoing = false;
        }
    }

    gtk_widget_destroy( GTK_WIDGET(InstallDialog) );
}

//--------------------------------------------------------------------------//
//  Create and display the un-installer dialog window
//--------------------------------------------------------------------------//
void
InstallManager::makeUnInstallDialogBox()
{
    UnInstallDialog = GTK_DIALOG(
        gtk_dialog_new_with_buttons( _("Module Un-installation"), NULL,
                                    GTK_DIALOG_MODAL, 
                                    GTK_STOCK_HELP, GTK_RESPONSE_HELP,
                                    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                    GTK_STOCK_OK, GTK_RESPONSE_OK, NULL ) );

    GtkWidget* vbox = UnInstallDialog->vbox;
    GtkWidget* table = gtk_table_new( 4, 1, FALSE );
    gtk_box_pack_start( GTK_BOX(vbox), table, TRUE, TRUE, 0 );
    gtk_container_set_border_width( GTK_CONTAINER(table), 10 );
    gtk_table_set_row_spacings( GTK_TABLE(table), 2 );
    gtk_table_set_col_spacings( GTK_TABLE(table), 5 );

    // Label above the text entry widget
    GtkWidget* label1 = gtk_label_new( _("Select a package to un-install and click OK:") );
    gtk_table_attach( GTK_TABLE(table), label1, 0, 1, 0, 1,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0 );
    gtk_label_set_justify( GTK_LABEL(label1), GTK_JUSTIFY_LEFT );
    gtk_misc_set_alignment( GTK_MISC(label1), 0.1, 0.5 );
    gtk_misc_set_padding( GTK_MISC(label1), 10, 0 );

    GtkWidget* sep = gtk_hseparator_new();
    gtk_table_attach( GTK_TABLE(table), sep, 0, 1, 2, 3,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (GTK_FILL), 0, 0 );

    // Status label and the text entry field
    UnInstallStatusLabel = makeInfoLabel( table, 3, 0 );

    GtkWidget* pkgEntry = gnome_entry_new( NULL );
    gtk_table_attach( GTK_TABLE(table), pkgEntry, 0, 1, 1, 2,
                     (GtkAttachOptions) (GTK_EXPAND | GTK_SHRINK | GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0 );
    gtk_container_set_border_width( GTK_CONTAINER(pkgEntry), 10 );

    UnInstallEntry = GTK_ENTRY( gnome_entry_gtk_entry( GNOME_ENTRY(pkgEntry) ) );

    gtk_widget_show_all( GTK_WIDGET(UnInstallDialog) );
    gtk_window_present( GTK_WINDOW(UnInstallDialog) );

    // Wait for user's response, might get more than one button click
    bool keepGoing = true;
    while ( keepGoing )
    {
        switch( gtk_dialog_run( UnInstallDialog ) )
        {
            case GTK_RESPONSE_OK:
                DoUnInstall();
                break;

            case GTK_RESPONSE_HELP:
                // TODO
                break;

            default: // cancel or delete event
                keepGoing = false;
        }
    }

    gtk_widget_destroy( GTK_WIDGET(UnInstallDialog) );
}

//--------------------------------------------------------------------------//
//  Download a file pointed by "src", save it as "dest".
//  Updates the progress bar as download progresses.
//--------------------------------------------------------------------------//
bool
InstallManager::downloadFile( const gchar* src, const gchar* dest ) const
{
    if ( Verbosity >= 3 )
        cout << "Downloading file from " << src << endl;

    WriteStatusMessage( "" );

    GnomeVFSURI* srcUri = gnome_vfs_uri_new( src );
    GnomeVFSURI* fileUri = gnome_vfs_uri_new( "file://" );
    GnomeVFSURI* destUri = gnome_vfs_uri_append_string( fileUri, dest );

    // If we're downloading something, say so in the status bar
    const gchar* scheme = gnome_vfs_uri_get_scheme( srcUri );
    bool notifyUser = ( strcmp(scheme, "file") != 0 );

    // TODO: don't call vfs when src is a local file
    if ( notifyUser )
        WriteStatusMessage( _("Downloading package...") );

    GnomeVFSResult result = gnome_vfs_xfer_uri( srcUri, destUri,
                            GNOME_VFS_XFER_DEFAULT,
                            GNOME_VFS_XFER_ERROR_MODE_ABORT,
                            GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
                            progressCallback, (gpointer)this );

    gnome_vfs_uri_unref( destUri );
    gnome_vfs_uri_unref( fileUri );
    gnome_vfs_uri_unref( srcUri );

    if ( result != GNOME_VFS_OK )
    {
        gchar* msg = g_strdup_printf( _("Error: %s"), 
                                      gnome_vfs_result_to_string( result ) );
        WriteStatusMessage( msg );
        g_free( msg );
        return false;
    }

    if ( notifyUser )
        WriteStatusMessage( _("Downloading is complete.") );

    return true;
}

//--------------------------------------------------------------------------//
//  Bring up install window
//--------------------------------------------------------------------------//
void
InstallManager::StartInstall( const string& path )
{
    makeInstallDialogBox();

    // Populate the entry widget if path is provided
    if ( !path.empty() )
    {
        gtk_entry_set_text( InstallPathEntry, path.c_str() );
        if ( Verbosity >= 3 )
            cout << "path = " << path << endl;
    }
}

//--------------------------------------------------------------------------//
//  Proceed with installation (OK button callback)
//--------------------------------------------------------------------------//
void
InstallManager::DoInstall()
{
    UpdateProgressBar( 0 );
    const gchar* path = gtk_entry_get_text( InstallPathEntry );
    if ( !path || !strcmp(path, "") )
        return;

    showPackageInfos( "", "", "", "" );

    // FIXME
    char tmpFile[] = "/tmp/gmodconfigXXXXXX";
    if ( mkstemp( tmpFile ) == -1 )
    {
        cout << "Could not create temp file" << endl;
        return;
    }

    // Download package file to /tmp and install it
    if ( downloadFile( path, tmpFile ) && installPackage( tmpFile ) )
        WriteStatusMessage( _("Installation succeeded") );

    if ( Verbosity < 2 ) // keep temp file around for debug purpose
        unlink( tmpFile );

    UpdateProgressBar( 0 );
}

//--------------------------------------------------------------------------//
//  Bring up uninstall window
//--------------------------------------------------------------------------//
void
InstallManager::StartUnInstall( const string& pkgName )
{
    makeUnInstallDialogBox();

    // Populate the entry widget if path is provided
    if ( !pkgName.empty() )
    {
        gtk_entry_set_text( UnInstallEntry, pkgName.c_str() );
        if ( Verbosity >= 3 )
            cout << "pkgName = " << pkgName << endl;
    }
}

//--------------------------------------------------------------------------//
//  Proceed with uninstallation (OK button callback)
//--------------------------------------------------------------------------//
void
InstallManager::DoUnInstall()
{
    const gchar* pkgName = gtk_entry_get_text( UnInstallEntry );
    if ( !pkgName || !strcmp(pkgName, "") )
        return;

    if ( uninstallPackage( pkgName ) )
        WriteStatusMessage( _("Installation succeeded") );
}

//--------------------------------------------------------------------------//
//  Checks if package is already installed
//--------------------------------------------------------------------------//
bool
InstallManager::IsPackageInstalled( const string& pkgName ) const
{
    return isPackageInstalled( pkgName );
}

//--------------------------------------------------------------------------//
//  Show package informations in the install mgr window
//--------------------------------------------------------------------------//
void
InstallManager::showPackageInfos( const string& name, const string& version,
                               const string& vendor, const string& summary )
{
    gtk_label_set_text( PkgNameLabel, name.c_str() );
    gtk_label_set_text( PkgVersionLabel, version.c_str() );
    gtk_label_set_text( PkgVendorLabel, vendor.c_str() );
    gtk_label_set_text( PkgSummaryLabel, summary.c_str() );
}

