//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  modconfmgr.cpp: modules.conf reader/writer
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "modconfmgr.h"
#include <gnome.h>
#include <stdexcept>
#include <fstream>
#include <sstream>
#include <iterator>
#include <cstdio>
#include <algorithm>
#include <vector>
#include <set>
using std::vector;
using std::set;

#include <iostream>
using std::cout;
using std::endl;

using namespace gmodcfg;

//--------------------------------------------------------------------------//
//  C'tor
//--------------------------------------------------------------------------//
ModulesConfManager::ModulesConfManager( const string& confFile, int verb )
        : Verbosity(verb), ModulesConfFile(confFile)
{
    if ( Verbosity )
        cout << "Using " << confFile << endl;
}

//--------------------------------------------------------------------------//
//  Load up the contents of the modules.conf file into CurrentOptionsMap
//--------------------------------------------------------------------------//
bool
ModulesConfManager::ReadModulesConf() throw()
{
    CurrentOptionsMap.clear();
    CurrentOptionsLineMap.clear();

    std::ifstream ifs( ModulesConfFile.c_str() );
    if ( !ifs )
        return false;

    int line = -1;
    string str;
    while ( getline( ifs, str ) )
    {
        ++line;
        if ( str.empty() )
            continue;

        std::istringstream istr( str );
        std::istream_iterator<string> it( istr ), itend;

        if ( (*it).at(0) == '#' )
            continue; // skip comments

        if ( Verbosity >= 2 )
            cout << "line #" << line << " = " << str << std::endl;

        // TODO: handle "alias" statements
        if ( *it != "options" )
            continue;  // only looking for options statements, for now

        string modName = *(++it);
        std::ostringstream oss;
        for( ++it; it != itend; ++it )
        {
            // Check options are well formed (no space around "=")
            int equalSign = (*it).find_first_of("=");
            if ( equalSign == string::npos )
                continue;
            oss << *it << " ";
        }

        CurrentOptionsMap[ modName ] = oss.str();
        CurrentOptionsLineMap[ modName ] = line; // keep track of line numbers
    }

    if ( Verbosity )
        cout << "Found " << CurrentOptionsMap.size() << " options" << endl;

    return true;
}

//--------------------------------------------------------------------------//
//  Returns the list of options associated to a module or "" if none.
//  Search first in ChangesMap, else in CurrentOptionsMap.
//--------------------------------------------------------------------------//
const string&
ModulesConfManager::GetOptions( const string& modName ) const
{
    StringStringMap_t::const_iterator it = ChangesMap.find( modName );
    if ( it != ChangesMap.end() )
        return (*it).second;
    it = CurrentOptionsMap.find( modName );
    if ( it != CurrentOptionsMap.end() )
        return (*it).second;
    return EmptyString;
}

//--------------------------------------------------------------------------//
//  Set (and overwrites) options of a module in ChangesMap. Remove entry
//  from ChangesMap if options is empty. Nothing written to file yet.
//--------------------------------------------------------------------------//
void
ModulesConfManager::SetOptions( const string& modName,
                                const string& options ) throw()
{
    set<string> newOpts;

    // Sanity check: make sure options string is well-formed
    std::istringstream istr( options );
    std::istream_iterator<string> is( istr ), isend;
    for( ; is != isend; ++is )
    {
        newOpts.insert( *is );
        int equalSign = (*is).find_first_of("=");
        if ( equalSign == string::npos )
            throw std::runtime_error(_("SetOptions called with bogus values") );
    }

    // Also compare provided options to the ones in CurrentOptionsMap.
    // Don't set anything if they're identical.
    StringStringMap_t::const_iterator mit = CurrentOptionsMap.find( modName );
    if ( mit != CurrentOptionsMap.end() )
    {
        set<string> oldOpts, diff;
        std::istringstream istr( (*mit).second );
        std::istream_iterator<string> is( istr ), isend;
        std::insert_iterator< set<string> > ii( oldOpts, oldOpts.begin() );
        std::copy( is, isend, ii );

        std::insert_iterator< set<string> > diffii( diff, diff.begin() );
        std::set_symmetric_difference( oldOpts.begin(), oldOpts.end(),
                                       newOpts.begin(), newOpts.end(), diffii );
        if ( diff.empty() )
        {
            if ( Verbosity )
                cout << "SetOptions: no changes" << endl;
            return;
        }
    }

    if ( options.empty() )
    {
        ChangesMap.erase( modName );

        if ( Verbosity >= 2 )
            cout << "SetOptions: discarding changes to module=\"" << modName
                 << "\"" << endl;
    }
    else
    {
        ChangesMap[ modName ] = options;

        if ( Verbosity >= 2 )
            cout << "SetOptions: update module=\"" << modName
                 << "\" options=\"" << options << "\"" << endl;
    }
}

//--------------------------------------------------------------------------//
//  Returns whether the options of a module have been modified
//--------------------------------------------------------------------------//
bool
ModulesConfManager::HasUnsavedChanges() const
{
    return ChangesMap.size() != 0;
}

//--------------------------------------------------------------------------//
//  Write out all the options that have been set with SetOptions
//--------------------------------------------------------------------------//
bool
ModulesConfManager::WriteAllOptions()
{
    if ( ChangesMap.empty() )
        return true; // nothing to write

    if ( Verbosity )
        cout << "Writing " << ModulesConfFile
             << " file...\nChanges in:" << endl;

    // Separate out existing options from new ones
    typedef map<int, string> IntStringMap_t;
    IntStringMap_t orderMap;
    vector<string> appendOpts;

    StringStringMap_t::const_iterator mit = ChangesMap.begin();
    StringStringMap_t::const_iterator mitend = ChangesMap.end();
    for( ; mit != mitend; ++mit )
    {
        const string& modName = (*mit).first;
        if ( Verbosity )
            cout << " " << modName;

        int line = getOptionLine( modName );
        if ( line == -1 )
            appendOpts.push_back( modName );
        else
            orderMap[ line ] = modName;
    }

    if ( Verbosity )
        cout << endl;

    // Backup modules.conf file
    rename( ModulesConfFile.c_str(), (ModulesConfFile + ".bak").c_str() );

    std::ifstream ifs( (ModulesConfFile + ".bak").c_str() );
    std::ofstream ofs( ModulesConfFile.c_str() );

    if ( !ofs )
        return false;

    // First update options that already exist in the file
    int currLine = 0;
    string str;
    IntStringMap_t::const_iterator it = orderMap.begin();
    IntStringMap_t::const_iterator itend = orderMap.end();
    for( ; it != itend; ++it )
    {
        int line = (*it).first;
        const string& modName = (*it).second;

        // The lines from "currLine" up to "line" are copied, intact
        for( int i=currLine; i<line; ++i )
        {
            getline( ifs, str );
            ofs << str << endl;
        }

        // Discard the current options line (do nothing with it)
        getline( ifs, str );

        // Replace it with the new options line (if any)
        const string& options = GetOptions( modName );
        if ( !options.empty() )
            ofs << "options " << modName << options << endl;

        currLine = line + 1;
    }

    // The remaining lines are then written out, intact
    while( getline( ifs, str ) )
        ofs << str << endl;

    // Now append the options that did not already exist
    vector<string>::const_iterator vit = appendOpts.begin();
    vector<string>::const_iterator vitend = appendOpts.end();
    for( ; vit != vitend; ++vit )
    {
        const string& modName = *vit;
        const string& options = GetOptions( modName );
        if ( !options.empty() )
            ofs << "\n# " << _("added by gmodconfig")
                << "\noptions " << modName << options << endl;
    }

    // Data has been written out, we can clear our map of changes
    ChangesMap.clear();

    // Need to refresh CurrentOptionsMap. Just re-read the file.
    ReadModulesConf();

    return true;
}

//--------------------------------------------------------------------------//
//  Returns the line # of an option, or -1 if none
//--------------------------------------------------------------------------//
int
ModulesConfManager::getOptionLine( const string& modName ) const
{
    StringIntMap_t::const_iterator it = CurrentOptionsLineMap.find( modName );
    if ( it == CurrentOptionsLineMap.end() )
        return -1;
    return (*it).second;
}

