//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  modinforeader.cpp: gather module informations from XML and modinfo
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "modinforeader.h"
#include "moduletypes.h"
#include "pipereader.h"
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libxml/parser.h>
#include <stdexcept>
#include <iostream>
#include <iterator>
#include <sstream>
#include "xsareader.h"
#include "xmlutils.h"

using namespace gmodcfg;
using namespace std;

//--------------------------------------------------------------------------//
//  D'tor
//--------------------------------------------------------------------------//
ModuleInfoReader::~ModuleInfoReader()
{
    ClearCache();
}

//--------------------------------------------------------------------------//
//  Helper function to parse a parameter definition in an XML tree
//  Returns: ptr to allocated Param object or NULL
//--------------------------------------------------------------------------//
static Param*
parseParamXml( xmlDoc* doc, xmlNode* node )
{
    if ( !node )
        return NULL;
    string name = XmlGetAttribute( node, "name" );
    string type = XmlGetAttribute( node, "type" );

    Param* p = new Param( name, type );
    p->SetValues( XmlGetAttribute( node, "values" ) );
    p->SetMin( XmlGetAttribute( node, "min" ) );
    p->SetMax( XmlGetAttribute( node, "max" ) );
    p->SetDefault( XmlGetAttribute( node, "default" ) );

    // Parse the description(s)
    for( xmlNode* iter = node->xmlChildrenNode; iter; iter = iter->next )
    {
        if ( !xmlStrcmp( iter->name, (const xmlChar*)"desc" ) )
        {
            string text = XmlGetContent( doc, iter );
            if ( text.empty() )
                continue;
            p->SetDescription( text, XmlGetAttribute( iter, "lang" ) );
        }
    }

    return p;
}

//--------------------------------------------------------------------------//
//  Helper function to parse a parameter group definition in an XML tree
//  Returns: ptr to allocated ParameterGroup object or NULL
//--------------------------------------------------------------------------//
static ParameterGroup*
parseParameterGroupXml( xmlDoc* doc, xmlNode* nodeGroup )
{
    if ( !nodeGroup )
        return NULL;
    string name = XmlGetAttribute( nodeGroup, "name" );
    ParameterGroup* pg = new ParameterGroup( name );

    // Below paramgroup is either param or desc
    xmlNode* iter = nodeGroup->xmlChildrenNode;
    for( ; iter; iter = iter->next )
    {
        if ( !xmlStrcmp( iter->name, (const xmlChar*)"param" ) )
        {
            Param* p = parseParamXml( doc, iter );
            pg->AddParam( p );
        }
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"desc" ) )
        {
            string text = XmlGetContent( doc, iter );
            if ( text.empty() )
                continue;
            pg->SetDescription( text, XmlGetAttribute( iter, "lang" ) );
        }
    }

    if ( pg->IsEmpty() )
    {
        delete pg;
        return NULL;
    }

    return pg;
}

//--------------------------------------------------------------------------//
//  Download an XSA file and parse it
//--------------------------------------------------------------------------//
static XsaData*
getXsaData( const string& src, int verbosity )
{
    char buf[] = "/tmp/gmodconfigXXXXXX";
    if ( mkstemp( buf ) == -1 )
        return 0;

    if ( verbosity )
        cout << "Getting XSA file from " << src << endl;

    GnomeVFSURI* srcUri = gnome_vfs_uri_new( src.c_str() );
    GnomeVFSURI* fileUri = gnome_vfs_uri_new( "file://" );
    GnomeVFSURI* destUri = gnome_vfs_uri_append_path( fileUri, buf );

    GnomeVFSResult result = gnome_vfs_xfer_uri( srcUri, destUri,
                            GNOME_VFS_XFER_DEFAULT,
                            GNOME_VFS_XFER_ERROR_MODE_ABORT,
                            GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
                            NULL, NULL );

    gnome_vfs_uri_unref( srcUri );
    gnome_vfs_uri_unref( fileUri );

    if ( result != GNOME_VFS_OK )
    {
        cout << _("Error reading the XSA file: ")
             << gnome_vfs_result_to_string( result ) << endl;
        return 0;
    }

    std::ifstream ifs( buf );
    XsaReader reader( verbosity );
    XsaData* data = reader.Read( ifs );
    unlink( buf );

    return data;
}

//--------------------------------------------------------------------------//
//  Read in module information from an XML stream
//  Returns: ptr to an allocated Module object
//--------------------------------------------------------------------------//
static Module*
readXmlStream( istream& is, int verbosity )
{
    if ( verbosity )
        cout << "Reading XML from stream..." << endl;

    xmlDoc* doc = XmlParseFromStream( is );
    xmlNode* node = xmlDocGetRootElement( doc );
    if ( !node )
        throw std::runtime_error("Empty XML document");

    // Traverse the document tree. First level must be <module>
    if ( xmlStrcmp( node->name, (const xmlChar *)"module" ) )
        throw std::runtime_error("Missing top level node \"module\"");

    string modName = XmlGetAttribute( node, "name" );
    if ( modName.empty() )
        throw std::runtime_error("Module node must have a name attribute");
 
    // Create a new Module
    Module* mod = new Module( modName );
    if ( verbosity )
        cout << "module name=" << mod->GetName() << endl;

    // 2nd level is either param, or paramgroup, or url, or shortdesc,
    // or longdesc
    node = node->xmlChildrenNode;
    for( ; node; node = node->next )
    {
        if ( !xmlStrcmp( node->name, (const xmlChar*)"param" ) )
        {
            Param* p = parseParamXml( doc, node );
            ParameterGroup* pg = new ParameterGroup("");
            pg->AddParam( p );
            mod->AddGroup( pg );
        }
        else if ( !xmlStrcmp( node->name, (const xmlChar*)"paramgroup" ) )
        {
            ParameterGroup* pg = parseParameterGroupXml( doc, node );
            mod->AddGroup( pg );
        }
        else if ( !xmlStrcmp( node->name, (const xmlChar*)"url" ) )
        {
            mod->SetURL( XmlGetContent( doc, node ) );
            if ( verbosity )
                cout << "url=" << mod->GetURL() << endl;
        }
        else if ( !xmlStrcmp( node->name, (const xmlChar*)"xsa" ) )
        {
            string xsaLoc = XmlGetContent( doc, node );
            if ( xsaLoc.empty() )
                continue;

            if ( verbosity )
                cout << "xsa=" << xsaLoc << endl;

            // Retrieve XSA file right now
            XsaData* data = getXsaData( xsaLoc, verbosity );
            mod->SetXsaURL( xsaLoc );
            mod->SetXsaData( data );
        }
        else if ( !xmlStrcmp( node->name, (const xmlChar*)"usage" ) )
        {
            string lang = XmlGetAttribute( node, "lang" );
            string text = XmlGetContent( doc, node );
            if ( text.empty() )
                continue;

            mod->SetUsage( text, lang );
            if ( verbosity )
                cout << "usage (" << lang << ")=" << mod->GetUsage( lang )
                     << endl;
        }
        else if ( !xmlStrcmp( node->name, (const xmlChar*)"about" ) )
        {
            string lang = XmlGetAttribute( node, "lang" );
            string text = XmlGetContent( doc, node );
            if ( text.empty() )
                continue;

            mod->SetAbout( text, lang );
            if ( verbosity )
                cout << "about (" << lang << ")=" << mod->GetAbout( lang )
                     << endl;
        }
    }

    xmlFreeDoc( doc );

    if ( verbosity )
        cout << "Done reading XML from stream." << endl;

    return mod;
}

//--------------------------------------------------------------------------//
//  Helper function to parse parameters info from the "old modinfo" output
//--------------------------------------------------------------------------//
static void
parseParamLineOldStyle( Module* mod, const string& line, int verbosity=0 )
{
    std::istringstream istr( line );
    std::istream_iterator<string> is( istr ), isend;
    ++is; // skip "parm:"

    string name = *is;

    // Perhaps no name is available...
    if ( name == "warning:" )
        return;

    ++is;
    string type = *is;
    if ( type.find_first_of( ',' ) != string::npos )
        type = type.substr( 0, type.length()-1 );

    // Now move on to "array" or "description"
    ++is;

    string minStr, maxStr;
    if ( (is != isend) && (*is == "array") )
    {
        type = "array";
        ++is; ++is; ++is;
        minStr = string( *is, 0, (*is).length()-1 ); // remove ","
        ++is; ++is; ++is;
        int rightParens = (*is).find_first_of( ')' );
        maxStr = string( *is, 0, rightParens ); // remove "),"
        ++is; // now we're on "description"
    }

    // Lookup this parameter. If it does not exist, create it
    Param* p = mod->Find( name );
    if ( !p )
    {
        p = new Param( name, type );
        p->SetMin( minStr );
        p->SetMax( maxStr );
        ParameterGroup* pg = new ParameterGroup("");
        pg->AddParam( p );
        mod->AddGroup( pg );
    }

    bool hasDesc = ( is != isend );

    if ( verbosity )
    {
        cout << name << " [" << type << "] ";
        if ( type == "array" )
            cout << "min=[" << minStr << "] max=[" << maxStr << "] ";
    }

    // We're overriding the parameter description (in English) only
    // if none are available already from the XML.
    if ( !p->GetDescription( "en" ).empty() )
        return;

    // If no description available from modinfo, use module name
    if ( !hasDesc )
    {
        p->SetDescription( name, "en" );
        return;
    }

    // Description is everything until the end of the line. We don't
    // need the istream_iterator anymore at this point.
    string desc = line.substr( istr.tellg(), line.length() );
    desc = desc.erase( desc.find_first_of( '"' ), 1 );
    int lastQuote = desc.find_last_of( '"' );
    if ( lastQuote != string::npos )
        desc = desc.erase( desc.find_last_of( '"' ), 1 );

    p->SetDescription( desc, "en" );

    if ( verbosity )
        cout << "desc=\"" << desc << "\"" << endl;
}

//--------------------------------------------------------------------------//
//  Helper function to parse parameters info from the "new modinfo" output
//--------------------------------------------------------------------------//
static void
parseParamLine( Module* mod, const string& line, int verbosity=0 )
{
    // Format: parm:   <name>:<desc>
    int nameStart = line.find_first_not_of( ' ', sizeof("parm:") );
    int nameEnd = line.find_first_of( ':', nameStart );

    string name, desc;
    if ( nameEnd != string::npos )
    {
        name = line.substr( nameStart, nameEnd-nameStart );

        int descStart = nameEnd + 1;
        int descEnd = line.length() - 1;
        if ( descStart < descEnd )
            desc = line.substr( descStart, descEnd-descStart+1 );
    }
    else
        name = line.substr( nameStart, line.length()-nameStart );

    // Lookup this parameter. If it does not exist, create it
    Param* p = mod->Find( name );
    if ( !p )
    {
        p = new Param( name, "string" ); // no type info in modinfo anymore...
        ParameterGroup* pg = new ParameterGroup("");
        pg->AddParam( p );
        mod->AddGroup( pg );
    }

    // We're overriding the parameter description (in English) only
    // if none are available already from the XML.
    if ( !p->GetDescription( "en" ).empty() )
        return;

    // If no description available from modinfo, use module name
    if ( desc.empty() )
    {
        p->SetDescription( name, "en" );
        return;
    }

    p->SetDescription( desc, "en" );
    
    if ( verbosity )
        cout << "name=\"" << name << "\" desc=\"" << desc << "\"" << endl;
}

//--------------------------------------------------------------------------//
//  Read in more informations from modinfo's output. This can also serve
//  as fallback in case no XML file is available for this module.
//  Module object will be built if not already.
//  Returns: 0 if failure
//--------------------------------------------------------------------------//
static Module*
readFromModinfo( const string& modName, Module* inmod, int verbosity=0 )
{
    if ( verbosity )
        cout << "Reading data from modinfo for module \""
             << modName << "\"" << endl;

    if ( modName.empty() )
    {
        if ( verbosity )
            cout << "Module has no name, bailing out..." << endl;
        return 0;
    }

    string cmd = MODINFO_DIR;
    cmd += "/modinfo " + modName;

    ifstream ifs;
    PipeReader pr( cmd, ifs );

    string line;
    getline( ifs, line );

    if ( line.empty() )
    {
        if ( verbosity )
            cout << "Could not find this module..." << endl;
        return 0;
    }

    // modinfo for Linux 2.4 and before starts its output with "filename:"
    // whereas modinfo for Linux 2.6 does not. This is used to distinguish
    // the 2 versions.
    bool oldmodinfo = ( line.compare( 0, 9, "filename:" ) == 0 );

    // Build a Module, if not built already
    Module* mod = inmod;
    if ( !mod )
        mod = new Module( modName );

    do
    {
        if ( verbosity >= 4 )
            cout << "modinfo> " << line << endl;

        int colon = line.find_first_of(":");
        if ( colon == string::npos )  // should not happen
            continue;

        // Parameter descriptions start with "parm:"
        if ( line.compare( 0, colon, "parm" ) == 0 )
        {
            if ( oldmodinfo )
                parseParamLineOldStyle( mod, line, verbosity );
            else
                parseParamLine( mod, line, verbosity );
        }

        // Module's short description
        else if ( line.compare( 0, colon, "description" ) == 0 )
        {
            int start = line.find_first_not_of( "\" ", colon+1 );
            int end = line.length()-1;
            if ( line.at(end) == '"' )
                --end;

            if ( mod->GetUsage().empty() && start != string::npos )
            {
                mod->SetUsage( line.substr( start, end-start+1 ), "en" );
                if ( verbosity )
                    cout << "description=\"" << mod->GetUsage() << "\"" << endl;
            }
        }

        // Authors
        else if ( line.compare( 0, colon, "author" ) == 0 )
        {
            int start = line.find_first_not_of( "\" ", colon+1 );
            int end = line.length()-1;
            if ( line.at(end) == '"' )
                --end;

            mod->SetAuthors( line.substr( start, end-start+1 ) );
            if ( verbosity )
                cout << "authors=\"" << mod->GetAuthors() << "\"" << endl;
        }

        // License
        else if ( line.compare( 0, colon, "license" ) == 0 )
        {
            int start = line.find_first_not_of( "\" ", colon+1 );
            int end = line.length()-1;
            if ( line.at(end) == '"' )
                --end;

            mod->SetLicense( line.substr( start, end-start+1 ) );
            if ( verbosity )
                cout << "license=\"" << mod->GetLicense() << "\"" << endl;
        }
    }
    while ( getline( ifs, line ) );

    if ( verbosity )
        cout << "Done reading from modinfo." << endl;

    return mod;
}

//--------------------------------------------------------------------------//
//  Main entry point
//  Assumption: a module whose name is "modName" exists
//  Returns: a new Module object if success, else NULL
//--------------------------------------------------------------------------//
const Module*
ModuleInfoReader::GetInfo( const string& modName, const string& lang )
{
    if ( Verbosity )
        cout << "\nGetting infos for module " << modName << "..." << endl;

    // Look up cache first
    Module* mod = getModule( modName );
    if ( mod )
    {
        if ( Verbosity )
            cout << "Found infos in cache" << endl;

        mod->SetLanguage( lang );
        return mod;
    }

    // Not found in cache. First try reading the associated XML file
    // (if any) in central directory
    string xmlFile = PACKAGE_DATA_DIR "/gmodconfig/xml/" + modName + ".xml";
    if ( Verbosity )
        cout << "Trying to access " << xmlFile << "..." << endl;

    ifstream ifs ( xmlFile.c_str() );
    if ( ifs.good() )
    {
        try
        {
            mod = readXmlStream( ifs, Verbosity );
        }
        catch( const std::exception& e )
        {
            if ( Verbosity )
                cout << "Exception caught in ModuleInfoReader: "
                     << e.what() << endl;
        }
    }

    // Parse modinfo's output to get more informations (or as fallback)
    // Module object will be built if not already.
    try
    {
        mod = readFromModinfo( modName, mod, Verbosity );
    }
    catch( const std::exception& e )
    {
        if ( Verbosity )
            cout << "Exception caught in ModuleInfoReader: "
                 << e.what() << endl;
    }

    // Update our cache
    addModule( modName, mod );

    if ( mod )
        mod->SetLanguage( lang );

    return mod;
}

//--------------------------------------------------------------------------//
//  Read from modinfo's output
//--------------------------------------------------------------------------//
Module*
ModuleInfoReader::ReadFromModinfo( const string& modName ) const
{
    Module* mod = 0;

    try
    {
        mod = readFromModinfo( modName, mod, Verbosity );
    }
    catch( const std::exception& e )
    {
        if ( Verbosity )
            cout << "Exception caught in ModuleInfoReader::ReadFromModinfo "
                 << e.what() << endl;
    }

    return mod;
}

//--------------------------------------------------------------------------//
//  Read from a stream
//--------------------------------------------------------------------------//
Module*
ModuleInfoReader::ReadFromStream( istream& is ) const
{
    Module* mod = 0;

    try
    {
        mod = readXmlStream( is, Verbosity );
    }
    catch( const std::exception& e )
    {
        if ( Verbosity )
            cout << "Exception caught in ModuleInfoReader: "
                 << e.what() << endl;
    }
    return mod;
}

//--------------------------------------------------------------------------//
//  Module infos cache helpers
//--------------------------------------------------------------------------//
Module*
ModuleInfoReader::getModule( const string& str ) const
{
    StringModuleMap_t::const_iterator it = ModuleCache.find( str );
    if ( it != ModuleCache.end() )
        return (*it).second;
    return 0;
}

void
ModuleInfoReader::addModule( const string& str, Module* newmod )
{
    if ( !newmod )
        return;
    if ( getModule( str ) )
        return;

    if ( Verbosity )
        cout << "Adding infos for " << str << " module to cache" << endl;

    ModuleCache[ str ] = newmod;
}

void
ModuleInfoReader::replaceModule( const string& str, Module* newmod )
{
    StringModuleMap_t::iterator it = ModuleCache.find( str );
    if ( it != ModuleCache.end() )
    {
        delete (*it).second;
        (*it).second = newmod;
    }
}

//--------------------------------------------------------------------------//
//  Throw away the cache
//--------------------------------------------------------------------------//
void
ModuleInfoReader::ClearCache()
{
    StringModuleMap_t::iterator it = ModuleCache.begin();
    StringModuleMap_t::iterator itend = ModuleCache.end();
    for( ; it != itend; ++it )
        delete (*it).second;
    ModuleCache.clear();
}

