//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  modinfowriter.cpp: writes out an XML description of a module
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "modinfowriter.h"
#include "moduletypes.h"
#include <iostream>
#include <cassert>

using namespace gmodcfg;
using namespace std;

//--------------------------------------------------------------------------//
//  Write out a parameter
//--------------------------------------------------------------------------//
static void
writeParam( ostream& os, const Param* p, const string& indent = "" )
{
    os << indent << "\n<param name=\"" << p->GetName()
       << "\" type=\"" << p->GetTypeString() << "\"";

    switch ( p->GetType() )
    {
        case PARAM_ARRAY:
        {
            os << " min=\"" << p->GetMin()
               << "\" max=\"" << p->GetMax() << "\"";
            break;
        }

        case PARAM_INT:
        case PARAM_LONG:
        case PARAM_SHORT:
        case PARAM_CHAR:
        case PARAM_STRING:
        {
            const string& values = p->GetValues();
            if ( !values.empty() )
                os << " values=\"" << p->GetValues() << "\"";
        }
        default:
            break;    
    }

    if ( !p->GetDefault().empty() )
        os << " default=\"" << p->GetDefault() << "\"";

    os << ">" << endl;

    StringStringMap_t::const_iterator dit = p->BeginDescription();
    StringStringMap_t::const_iterator ditend = p->EndDescription();
    for( ; dit != ditend; ++dit )
    {
        const string& str = (*dit).second;
        if ( !str.empty() )
            os << indent << "  <desc xml:lang=\"" << (*dit).first << "\">"
               << str << "</desc>" << endl;
    }

    os << indent << "</param>" << endl;
}

//--------------------------------------------------------------------------//
//  Write out a parameter group
//--------------------------------------------------------------------------//
static void
writeParamGroup( ostream& os, const ParameterGroup* pg )
{
    assert( !pg->IsEmpty() );

    if ( pg->GetNumParams() > 1 )
    {
        os << "<paramgroup>" << endl;
        ParameterVector_t::const_iterator it = pg->BeginParams();
        ParameterVector_t::const_iterator itend = pg->EndParams();
        for( ; it != itend; ++it )
            writeParam( os, *it, "  " );
        os << "</paramgroup>" << endl;
    }
    else
        writeParam( os, *( pg->BeginParams() ) );
}

//--------------------------------------------------------------------------//
//  Main entry point
//--------------------------------------------------------------------------//
bool
ModuleInfoWriter::WriteToStream( const Module* mod, ostream& os ) const
{
    if ( Verbosity )
        cout << "Writing module infos to XML..." << endl;

    os << "<?xml version=\"1.0\"?>" << endl;
    os << "<!-- generated by gmodconfig's module XML editor -->" << endl;
    os << "<!-- visit http://gmodconfig.sourceforge.net/ for informations -->"
       << endl;

    // Module name, URL, XSA
    os << "\n<module name=\"" << mod->GetName() << "\">" << endl;
    os << "<url>" << mod->GetURL() << "</url>" << endl;
    os << "<xsa>" << mod->GetXsaURL() << "</xsa>\n" << endl;

    // Usage
    StringStringMap_t::const_iterator uit = mod->BeginUsage();
    StringStringMap_t::const_iterator uitend = mod->EndUsage();
    for( ; uit != uitend; ++uit )
    {
        const string& str = (*uit).second;
        if ( !str.empty() )
            os << "<usage xml:lang=\"" << (*uit).first << "\">"
               << str << "</usage>" << endl;
    }
    os << endl;

    // About
    StringStringMap_t::const_iterator ait = mod->BeginAbout();
    StringStringMap_t::const_iterator aitend = mod->EndAbout();
    for( ; ait != aitend; ++ait )
    {
        const string& str = (*ait).second;
        if ( !str.empty() )
            os << "<about xml:lang=\"" << (*ait).first << "\">"
               << str << "</about>" << endl;
    }
    os << endl;

    ParameterDB_t::const_iterator pit = mod->BeginGroups();
    ParameterDB_t::const_iterator pitend = mod->EndGroups();
    for( ; pit != pitend; ++pit )
        writeParamGroup( os, *pit );

    if ( Verbosity >= 2 )
        cout << "There are " << mod->GetNumGroups()
             << " parameter groups" << endl;

    os << "\n</module>" << endl;

    if ( Verbosity )
        cout << "Done writing module infos to XML." << endl;
}

