//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  moduletypes.cpp: Module and Parameter classes
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "moduletypes.h"
#include <algorithm>

using namespace gmodcfg;

// Keep the types in the same order as in moduletypes.h
static const char* ParamTypeNames[] =
  { "bool", "array", "int", "long", "byte", "short", "char", "string", NULL };

const char**
Param::GetTypeList() { return ParamTypeNames; }

//--------------------------------------------------------------------------//
//  Param c'tor
//--------------------------------------------------------------------------//
Param::Param( const string& name, const string& type )
    : Name(name), Type(PARAM_UNKNOWN)
{
    SetType( type );
}

//--------------------------------------------------------------------------//
//  Set type based on a string
//--------------------------------------------------------------------------//
void
Param::SetType( const string& type )
{
    for( unsigned int i=0; ParamTypeNames[i]; ++i )
    {
        if ( strcmp( type.c_str(), ParamTypeNames[i] ) == 0 )
        {
            Type = (ParamType) i ;
            break;
        }
    }
}

//--------------------------------------------------------------------------//
//  Return a string representation of a type
//--------------------------------------------------------------------------//
string
Param::GetTypeString() const
{
    return ParamTypeNames[ Type ];
}

//--------------------------------------------------------------------------//
//  Get/Set description, in a certain language
//--------------------------------------------------------------------------//
const string&
Param::GetDescription( const string& lang ) const
{
    StringStringMap_t::const_iterator it = DescMap.find( lang );
    if ( it != DescMap.end() )
        return (*it).second;
    return EmptyString;
}

void
Param::SetDescription( const string& desc, const string& lang )
{
    DescMap[ lang ] = desc;
}

//--------------------------------------------------------------------------//
//  Helper functor to delete things
//--------------------------------------------------------------------------//
struct deleteFunctor
{
    template<typename Stuff>
    void operator()( const Stuff* p ) const { delete p; }
};

//--------------------------------------------------------------------------//
//  ParameterGroup d'tor
//--------------------------------------------------------------------------//
ParameterGroup::~ParameterGroup()
{
    std::for_each( Params.begin(), Params.end(), deleteFunctor() );
}

//--------------------------------------------------------------------------//
//  Get/Set the description of a ParameterGroup
//--------------------------------------------------------------------------//
const string&
ParameterGroup::GetDescription( const string& lang ) const
{
    StringStringMap_t::const_iterator it = DescMap.find( lang );
    if ( it != DescMap.end() )
        return (*it).second;
    return EmptyString;
}

void
ParameterGroup::SetDescription( const string& desc, const string& lang )
{
    DescMap[ lang ] = desc;
}

//--------------------------------------------------------------------------//
//  Module d'tor
//--------------------------------------------------------------------------//
Module::~Module()
{
    std::for_each( Groups.begin(), Groups.end(), deleteFunctor() );
}

//--------------------------------------------------------------------------//
//  Add a new ParameterGroup
//--------------------------------------------------------------------------//
void
Module::AddGroup( ParameterGroup* pg )
{
    if ( pg && !pg->IsEmpty() )
        Groups.push_back( pg );
}

//--------------------------------------------------------------------------//
//  Get/Set Usage and About
//--------------------------------------------------------------------------//
const string&
Module::GetUsage( const string& lang ) const
{
    StringStringMap_t::const_iterator it =
        UsageMap.find( ( lang.empty() ? Language : lang ) );
    if ( it != UsageMap.end() )
        return (*it).second;
    return EmptyString;
}

void
Module::SetUsage( const string& str, const string& lang )
{
    UsageMap[ ( lang.empty() ? Language : lang ) ] = str;
}

const string&
Module::GetAbout( const string& lang ) const
{
    StringStringMap_t::const_iterator it =
        AboutMap.find( ( lang.empty() ? Language : lang ) );
    if ( it != AboutMap.end() )
        return (*it).second;
    return EmptyString;
}

void
Module::SetAbout( const string& str, const string& lang )
{
    AboutMap[ ( lang.empty() ? Language : lang ) ] = str;
}

//--------------------------------------------------------------------------//
//  Find by name
//--------------------------------------------------------------------------//
struct paramNameFunctor
{
    string Name;
    paramNameFunctor( const string& str ) : Name(str) {}
    bool operator()( const Param* p ) const { return p->GetName() == Name; }
};

Param*
Module::Find( const string& name ) const
{
    ParameterDB_t::const_iterator it = BeginGroups();
    ParameterDB_t::const_iterator itend = EndGroups();
    for( ; it != itend; ++it )
    {
        const ParameterGroup* pg = *it;
        ParameterVector_t::const_iterator pit = find_if( pg->BeginParams(),
                                    pg->EndParams(), paramNameFunctor(name) );
        if ( pit != pg->EndParams() )
            return *pit;
    }
    return 0;
}

