//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  moduletypes.h: Module and Param classes
//--------------------------------------------------------------------------//

#ifndef moduletypes_h
#define moduletypes_h

#include <string>
using std::string;

#include <vector>
#include <map>

namespace gmodcfg {

enum ParamType {
    PARAM_BOOL, PARAM_ARRAY, PARAM_INT, PARAM_LONG, PARAM_BYTE,
    PARAM_SHORT, PARAM_CHAR, PARAM_STRING, PARAM_UNKNOWN
};

typedef std::map<string, string> StringStringMap_t;

//--------------------------------------------------------------------------//
//  A module parameter
//--------------------------------------------------------------------------//
class Param
{
        string Name, Values, Min, Max, DefaultVal, EmptyString;
        ParamType Type;
        StringStringMap_t DescMap;

    public:
        Param( const string& name, const string& type );

        static const char** GetTypeList();

        const string& GetName() const { return Name; }
        void SetType( const string& );
        ParamType GetType() const { return Type; }
        string GetTypeString() const;
        const string& GetDefault() const { return DefaultVal; }
        void SetDefault( const string& val ) { DefaultVal = val; }

        void SetDescription( const string& desc, const string& lang );
        const string& GetDescription( const string& lang ) const;

        // i18n descriptions iterators (keyed on language)
        StringStringMap_t::const_iterator BeginDescription() const
                                                    { return DescMap.begin(); }
        StringStringMap_t::const_iterator EndDescription() const
                                                    { return DescMap.end(); }

        // For arrays
        const string& GetMin() const { return Min; }
        const string& GetMax() const { return Max; }
        void SetMin( const string& val ) { Min = val; }
        void SetMax( const string& val ) { Max = val; }

        // For int or strings
        const string& GetValues() const { return Values; }
        void SetValues( const string& val ) { Values = val; }
};

typedef std::vector<Param*> ParameterVector_t;

//--------------------------------------------------------------------------//
//  A class representing a group of parameters
//--------------------------------------------------------------------------//
class ParameterGroup
{
        string Name, EmptyString;
        ParameterVector_t Params;
        StringStringMap_t DescMap;

        ParameterGroup( const ParameterGroup& );
        ParameterGroup& operator=( const ParameterGroup& );

    public:
        explicit ParameterGroup( const string& name ) : Name(name) {}
        ~ParameterGroup();

        const string& GetName() const { return Name; }

        void SetDescription( const string& desc, const string& lang );
        const string& GetDescription( const string& lang ) const;

        // i18n descriptions iterators (keyed on language)
        StringStringMap_t::const_iterator BeginDescription() const
                                                    { return DescMap.begin(); }
        StringStringMap_t::const_iterator EndDescription() const
                                                    { return DescMap.end(); }

        ParameterVector_t::const_iterator BeginParams() const
                                          { return Params.begin(); }
        ParameterVector_t::const_iterator EndParams() const
                                          { return Params.end(); }
        unsigned int GetNumParams() const { return Params.size(); }
        bool IsEmpty() const { return Params.empty(); }

        void AddParam( Param* p ) { Params.push_back( p ); }
};

typedef std::vector<ParameterGroup*> ParameterDB_t;

//--------------------------------------------------------------------------//
//  A class holding a kernel module's informations
//--------------------------------------------------------------------------//
class XsaData;

class Module
{
        string Name, Url, XsaUrl, Authors, License, Language, EmptyString;
        StringStringMap_t UsageMap, AboutMap;
        XsaData* XsaInfos;
        ParameterDB_t Groups;

        Module( const Module& );
        Module& operator=( const Module& );

    public:
        Module( const string& name ) : Name(name), Language("en"),
                                       XsaInfos(0) {}
        ~Module();

        void SetName( const string& str ) { Name = str; }
        void SetURL( const string& str ) { Url = str; }
        void SetXsaURL( const string& str ) { XsaUrl = str; }
        void SetXsaData( XsaData* data ) { XsaInfos = data; }
        void SetUsage( const string& str, const string& lang );
        void SetAbout( const string& str, const string& lang );
        void SetAuthors( const string& str ) { Authors = str; }
        void SetLicense( const string& str ) { License = str; }
        void SetLanguage( const string& str ) { Language = str; }
        void AddGroup( ParameterGroup* );

        const string& GetName() const { return Name; }
        const string& GetURL() const { return Url; }
        const string& GetXsaURL() const { return XsaUrl; }
        XsaData* GetXsaData() const { return XsaInfos; }
        const string& GetUsage( const string& lang="" ) const;
        const string& GetAbout( const string& lang="" ) const;
        const string& GetAuthors() const { return Authors; }
        const string& GetLicense() const { return License; }
        const string& GetLanguage() const { return Language; }

        // i18n usage and "about" iterators (keyed on language)
        StringStringMap_t::const_iterator BeginUsage() const
                                                    { return UsageMap.begin(); }
        StringStringMap_t::const_iterator EndUsage() const
                                                    { return UsageMap.end(); }
        StringStringMap_t::const_iterator BeginAbout() const
                                                    { return AboutMap.begin(); }
        StringStringMap_t::const_iterator EndAbout() const
                                                    { return AboutMap.end(); }

        // ParameterGroup's iterators
        ParameterDB_t::const_iterator BeginGroups() const
                                                    { return Groups.begin(); }
        ParameterDB_t::const_iterator EndGroups() const { return Groups.end(); }
        unsigned int GetNumGroups() const { return Groups.size(); }
        Param* Find( const string& ) const;
};

};

#endif
