//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  modxmlmgr.cpp: Module XML assistant
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "modxmlmgr.h"
#include "moduletypes.h"
#include "modinforeader.h"
#include "modinfowriter.h"
#include "editinterface.h"
#include "support.h"
#include "treeviewmgr.h"

#include <fstream>
#include <iostream>
#include <set>

using namespace gmodcfg;
using namespace std;

enum { MOD_PARAM_NAME=0, MOD_PARAM_TYPE, MOD_PARAM_VALUES, MOD_PARAM_DEFAULT,
       MOD_PARAM_MIN, MOD_PARAM_MAX, MOD_PARAM_MODINFO, MOD_PARAM_DESC,
       MOD_NUM_COLUMNS };

static gchar* columnNames[] = { _("Name"), _("Type"), _("Description") };
static int columnsEnums[] = { MOD_PARAM_NAME, MOD_PARAM_TYPE, MOD_PARAM_MODINFO };

//--------------------------------------------------------------------------//
//  Get/Set helpers for GtkTextView's
//  Note: need to call g_free() on what textviewGetText() returns
//--------------------------------------------------------------------------//
static gchar*
textviewGetText( GtkTextView* widget )
{
    GtkTextBuffer* buf = gtk_text_view_get_buffer( widget );
    GtkTextIter start, end;
    gtk_text_buffer_get_bounds( buf, &start, &end );
    return gtk_text_iter_get_text( &start, &end );
}

static void
textviewSetText( GtkTextView* widget, const gchar* text )
{
    if ( !text )
        return;
    GtkTextBuffer* buf = gtk_text_view_get_buffer( widget );
    gtk_text_buffer_set_text( buf, text, -1 );
}

//--------------------------------------------------------------------------//
//  Treeview manager for parameters
//--------------------------------------------------------------------------//
class gmodcfg::ParamTreeViewManager : public TreeViewManager
{
        const Module* ModXmlData;
        const Module* ModInfoData;
        GtkLabel *ParamNameLabel;
        GtkEntry *ParamTypeEntry, *ParamValuesEntry, *ParamDefaultEntry,
                 *ParamMinEntry, *ParamMaxEntry;
        GtkTextView *ParamModinfoTextView, *ParamDescTextView;
        int Verbosity;
        set<string> Entries;

    public:
        ParamTreeViewManager( GtkTreeView* treeView, GtkLabel* name,
                GtkEntry* type, GtkEntry* values, GtkEntry* defaultVal,
                GtkEntry* minVal, GtkEntry* maxVal, GtkTextView* modinfo,
                GtkTextView* desc, int verbosity=0 )
            : TreeViewManager(treeView), ParamNameLabel(name),
              ParamTypeEntry(type), ParamValuesEntry(values),
              ParamDefaultEntry(defaultVal),
              ParamMinEntry(minVal), ParamMaxEntry(maxVal),
              ParamModinfoTextView(modinfo), ParamDescTextView(desc),
              ModXmlData(0), ModInfoData(0), Verbosity(verbosity) {}

        void SetModXmlData( Module* mod ) { ModXmlData = mod; }
        void SetModInfoData( Module* mod ) { ModInfoData = mod; }

        void ShowParameter( GtkTreeIter* );
        void CheckType();
        void PopulateListStore();

        // Insert/Remove param names (to avoid duplicates in treeview)
        bool InsertParameter( const gchar* );
        void RemoveParameter( const gchar* );

    private:
        virtual void createListStore();
        virtual void initTreeView();
        virtual void moveListItem( GtkTreeIter* from, GtkTreeIter* to );
        virtual void setListEntry( GtkTreeIter* );
};

//--------------------------------------------------------------------------//
//  Display details of a parameter
//--------------------------------------------------------------------------//
void
ParamTreeViewManager::ShowParameter( GtkTreeIter* iter )
{
    GtkTreeModel* model = GTK_TREE_MODEL( ListStore );
    gchar *name, *type, *values, *defaultVal, *minVal, *maxVal, *modinfo, *desc;
    gtk_tree_model_get( model, iter,
                        MOD_PARAM_NAME, &name,
                        MOD_PARAM_TYPE, &type,
                        MOD_PARAM_VALUES, &values,
                        MOD_PARAM_DEFAULT, &defaultVal,
                        MOD_PARAM_MIN, &minVal,
                        MOD_PARAM_MAX, &maxVal,
                        MOD_PARAM_MODINFO, &modinfo,
                        MOD_PARAM_DESC, &desc,
                        -1 );

    gtk_label_set_text( ParamNameLabel, name );
    gtk_entry_set_text( ParamTypeEntry, type );
    gtk_entry_set_text( ParamValuesEntry, values );
    gtk_entry_set_text( ParamDefaultEntry, defaultVal );
    gtk_entry_set_text( ParamMinEntry, minVal );
    gtk_entry_set_text( ParamMaxEntry, maxVal );
    textviewSetText( ParamModinfoTextView, modinfo );
    textviewSetText( ParamDescTextView, desc );

    CheckType();
}

//--------------------------------------------------------------------------//
//  Clear and make fields editable or non-editable based on type
//--------------------------------------------------------------------------//
void
ParamTreeViewManager::CheckType()
{
    const gchar* type = gtk_entry_get_text( ParamTypeEntry );
    if ( !type )
        return;

    // Min/Max entries are editable only for an array type
    if ( !strcmp(type, "array") )
    {
        gtk_editable_set_editable( GTK_EDITABLE(ParamMinEntry), TRUE );
        gtk_editable_set_editable( GTK_EDITABLE(ParamMaxEntry), TRUE );
    }
    else
    {
        gtk_editable_set_editable( GTK_EDITABLE(ParamMinEntry), FALSE );
        gtk_editable_set_editable( GTK_EDITABLE(ParamMaxEntry), FALSE );
        gtk_entry_set_text( ParamMinEntry, "" );
        gtk_entry_set_text( ParamMaxEntry, "" );
    }

    // Values entry is editable only for int, long, short and string types
    if ( !strcmp(type, "int") || !strcmp(type, "long")
       || !strcmp(type, "short") || !strcmp(type, "string") )
    {
        gtk_editable_set_editable( GTK_EDITABLE(ParamValuesEntry), TRUE );
    }
    else
    {
        gtk_editable_set_editable( GTK_EDITABLE(ParamValuesEntry), FALSE );
        gtk_entry_set_text( ParamValuesEntry, "" );
    }
}

//--------------------------------------------------------------------------//
//  Populate the list store. Get parameters from XML data if available,
//  else fallback on modinfo's output.
//--------------------------------------------------------------------------//
void
ParamTreeViewManager::PopulateListStore()
{
    // Sanity check
    if ( !ModXmlData )
    {
        cout << "ERROR: ModXmlData is null" << endl;
        return;
    }
 
    // Empty the list store first
    gtk_list_store_clear( ListStore );
 
    // Debug
    if ( Verbosity >= 2 )
    {
        if ( !ModInfoData )
            cout << "ModInfoData is null" << endl;
    }

    GtkTreeIter iter;
    ParameterDB_t::const_iterator it, itend;
   
    if ( ModXmlData->GetNumGroups() != 0 )
    {
        it = ModXmlData->BeginGroups();
        itend = ModXmlData->EndGroups();
    }
    else if ( ModInfoData )
    {
        it = ModInfoData->BeginGroups();
        itend = ModInfoData->EndGroups();
    }
    else
        return;

    for( ; it != itend; ++it )
    {
        const ParameterGroup* pg = *it;

        // FIXME: add support for parameter groups
        const Param* px = *(pg->BeginParams());

        // Retrieve corresponding Param data from modinfo
        const Param* pmi = ( ModInfoData ? ModInfoData->Find( px->GetName() )
                                         : 0 );
        gtk_list_store_append( ListStore, &iter );
        gtk_list_store_set( ListStore, &iter,
                            MOD_PARAM_NAME, px->GetName().c_str(),
                            MOD_PARAM_TYPE, px->GetTypeString().c_str(),
                            MOD_PARAM_VALUES, px->GetValues().c_str(),
                            MOD_PARAM_DEFAULT, px->GetDefault().c_str(),
                            MOD_PARAM_MIN, px->GetMin().c_str(),
                            MOD_PARAM_MAX, px->GetMax().c_str(),
                            MOD_PARAM_MODINFO,
                                ( pmi ? pmi->GetDescription("en").c_str()
                                      : _("Description not available") ),
                            -1 );

        // Only fill in user entry for parameter description if data was
        // read from XML data ( no point in duplicating modinfo's description)
        if ( ModXmlData->GetNumGroups() != 0 )
            gtk_list_store_set( ListStore, &iter,
                            MOD_PARAM_DESC, px->GetDescription("en").c_str(),
                            -1 );
    }
}

//--------------------------------------------------------------------------//
//  Insert/Remove param names (to avoid duplicates in treeview)
//--------------------------------------------------------------------------//
bool
ParamTreeViewManager::InsertParameter( const gchar* name )
{
    pair< set<string>::iterator, bool > p = Entries.insert( name );
    return p.second;
}

void
ParamTreeViewManager::RemoveParameter( const gchar* name )
{
    Entries.erase( name );
}

//--------------------------------------------------------------------------//
//  Build the list store for parameters
//--------------------------------------------------------------------------//
void
ParamTreeViewManager::createListStore()
{
    ListStore = gtk_list_store_new( MOD_NUM_COLUMNS, G_TYPE_STRING,
                                    G_TYPE_STRING, G_TYPE_STRING,
                                    G_TYPE_STRING, G_TYPE_STRING,
                                    G_TYPE_STRING, G_TYPE_STRING,
                                    G_TYPE_STRING );
}

//--------------------------------------------------------------------------//
//  Create columns in the tree view
//--------------------------------------------------------------------------//
void
ParamTreeViewManager::initTreeView()
{
    // Create only 3 columns in treeview: name, type, modinfo's description
    for( int i=0; i<3; ++i )
    {
        GtkCellRenderer* renderer = gtk_cell_renderer_text_new();
        GtkTreeViewColumn* column =
            gtk_tree_view_column_new_with_attributes( columnNames[i],
                                     renderer, "text", columnsEnums[i], NULL );
        gtk_tree_view_append_column( TreeView, column );
    }
}

//--------------------------------------------------------------------------//
//  Move parameter data from one place to another in the list store
//--------------------------------------------------------------------------//
void
ParamTreeViewManager::moveListItem( GtkTreeIter* from, GtkTreeIter* to )
{
    GtkTreeModel* model = GTK_TREE_MODEL( ListStore );
    gchar *name, *type, *values, *defaultVal, *minVal, *maxVal, *modinfo, *desc;
    gtk_tree_model_get( model, from,
                        MOD_PARAM_NAME, &name,
                        MOD_PARAM_TYPE, &type,
                        MOD_PARAM_VALUES, &values,
                        MOD_PARAM_DEFAULT, &defaultVal,
                        MOD_PARAM_MIN, &minVal,
                        MOD_PARAM_MAX, &maxVal,
                        MOD_PARAM_MODINFO, &modinfo,
                        MOD_PARAM_DESC, &desc,
                        -1 );
    gtk_list_store_set( ListStore, to,
                        MOD_PARAM_NAME, name,
                        MOD_PARAM_TYPE, type,
                        MOD_PARAM_VALUES, values,
                        MOD_PARAM_DEFAULT, defaultVal,
                        MOD_PARAM_MIN, minVal,
                        MOD_PARAM_MAX, maxVal,
                        MOD_PARAM_MODINFO, modinfo,
                        MOD_PARAM_DESC, desc,
                        -1 );
    gtk_list_store_remove( ListStore, from );
}

//--------------------------------------------------------------------------//
//  Update parameter data in the list store
//--------------------------------------------------------------------------//
void
ParamTreeViewManager::setListEntry( GtkTreeIter* iter )
{
    const gchar* name = gtk_label_get_text( ParamNameLabel );
    const gchar* type = gtk_entry_get_text( ParamTypeEntry );
    const gchar* values = gtk_entry_get_text( ParamValuesEntry );
    const gchar* defaultVal = gtk_entry_get_text( ParamDefaultEntry );
    const gchar* minVal = gtk_entry_get_text( ParamMinEntry );
    const gchar* maxVal = gtk_entry_get_text( ParamMaxEntry );

    // Check type
    if ( !strcmp(type, "") )
        return;
    if ( !strcmp(type, "array")
         && ( !strcmp(minVal, "") || !strcmp(maxVal, "") ) )
        return;

    gchar* modinfo = textviewGetText( ParamModinfoTextView );
    gchar* desc = textviewGetText( ParamDescTextView );

    gtk_list_store_set( ListStore, iter,
                        MOD_PARAM_NAME, name,
                        MOD_PARAM_TYPE, type,
                        MOD_PARAM_VALUES, values,
                        MOD_PARAM_DEFAULT, defaultVal,
                        MOD_PARAM_MIN, minVal,
                        MOD_PARAM_MAX, maxVal,
                        MOD_PARAM_MODINFO, modinfo,
                        MOD_PARAM_DESC, desc,
                        -1 );
    g_free( modinfo );
    g_free( desc );
}

//--------------------------------------------------------------------------//
//  Module XML druid
//--------------------------------------------------------------------------//
ModuleXmlManager::ModuleXmlManager( GtkWidget* appWidget, int verbosity )
    : Assistant(0), Druid(0), ModuleXmlInfos(0), ModuleModinfos(0),
      Verbosity(verbosity), TreeViewMgr(0), TreeView(0), ModinfoButton(0),
      ModNameEntry(0), ModUrlEntry(0), ModXsaEntry(0), ModUsageEntry(0),
      ModUsageModinfoTextView(0), ModAboutTextView(0),
      ParamNameLabel(0), ParamTypeEntry(0), ParamValuesEntry(0),
      ParamMinEntry(0), ParamMaxEntry(0), ParamModinfoTextView(0),
      ParamDescTextView(0), Initialized(false)
{
    StatusBar = GNOME_APPBAR( lookup_widget( appWidget, "editappbar" ) );
    Assistant = create_modwindow();
    Druid = GNOME_DRUID( lookup_widget( Assistant, "moddruid" ) );
    gtk_widget_show( Assistant );
}

ModuleXmlManager::~ModuleXmlManager()
{
    cleanUp();
    delete TreeViewMgr;
}

//--------------------------------------------------------------------------//
//  Clean up
//--------------------------------------------------------------------------//
void
ModuleXmlManager::cleanUp()
{
    delete ModuleXmlInfos;
    delete ModuleModinfos;
    ModuleXmlInfos = ModuleModinfos = 0;
}

//--------------------------------------------------------------------------//
//  Treeview's "changed" callback
//--------------------------------------------------------------------------//
static void
selectionChangedCallback( GtkTreeSelection* select, gpointer data )
{
    GtkTreeView* treeView = gtk_tree_selection_get_tree_view( select );
    GtkTreeModel* model = gtk_tree_view_get_model( treeView );
    ParamTreeViewManager* treeMgr = (ParamTreeViewManager*)data;
    GtkTreeIter iter;
    if ( gtk_tree_selection_get_selected( select, &model, &iter ) )
        treeMgr->ShowParameter( &iter );
}

//--------------------------------------------------------------------------//
//  Init (but don't populate) the druid
//--------------------------------------------------------------------------//
void
ModuleXmlManager::initDruid()
{
    if ( Initialized )
        return;

    ModinfoButton =
        lookup_widget( Assistant, "moddruid_header_modinfobutton" );
    ModNameEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_header_name" ) );
    ModUrlEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_header_url" ) );
    ModXsaEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_header_xsa" ) );
    ModUsageEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_header_usage" ) );
    ModUsageModinfoTextView =GTK_TEXT_VIEW( lookup_widget( Assistant,
                                            "moddruid_header_usage_modinfo" ) );
    ModAboutTextView =
        GTK_TEXT_VIEW( lookup_widget( Assistant, "moddruid_header_about" ) );

    ParamNameLabel =
        GTK_LABEL( lookup_widget( Assistant, "moddruid_param_name" ) );
    ParamTypeEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_param_type" ) );
    ParamValuesEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_param_values" ) );
    ParamDefaultEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_param_default" ) );
    ParamMinEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_param_min" ) );
    ParamMaxEntry =
        GTK_ENTRY( lookup_widget( Assistant, "moddruid_param_max" ) );
    ParamModinfoTextView =
        GTK_TEXT_VIEW( lookup_widget( Assistant, "moddruid_param_modinfo" ) );
    ParamDescTextView =
        GTK_TEXT_VIEW( lookup_widget( Assistant, "moddruid_param_desc" ) );

    // Add pulldown list to type combo widget
    GtkCombo* typeCombo =
        GTK_COMBO( lookup_widget( Assistant, "moddruid_param_type_combo" ) );

    const char** typeList = Param::GetTypeList();
    GList* vallist = NULL;
    for( int i=0; typeList[i]; ++i )
        vallist = g_list_append( vallist, (gpointer)typeList[i] );

    gtk_combo_set_popdown_strings( typeCombo, vallist );
    g_list_free( vallist );

    // Set up the treeview manager
    TreeView = GTK_TREE_VIEW( lookup_widget( Assistant,
                                            "moddruid_param_treeview" ) );
    delete TreeViewMgr;
    TreeViewMgr = new ParamTreeViewManager( TreeView, ParamNameLabel,
                    ParamTypeEntry, ParamValuesEntry, ParamDefaultEntry,
                    ParamMinEntry, ParamMaxEntry, ParamModinfoTextView,
                    ParamDescTextView, Verbosity );
    TreeViewMgr->InitListStore();

    // Install a callback to detect clicks on the treeview
    GtkTreeSelection* select = gtk_tree_view_get_selection( TreeView );
    g_signal_connect( G_OBJECT( select ), "changed",
                      G_CALLBACK( selectionChangedCallback ),
                      (gpointer)TreeViewMgr );

    Initialized = true;
}

//--------------------------------------------------------------------------//
//  Populate headers and parameters infos (page 2 and 3 of druid)
//  Don't overwrite existing header infos
//--------------------------------------------------------------------------//
void
ModuleXmlManager::populateDruid()
{
    if ( !strcmp( gtk_entry_get_text( ModNameEntry ), "" ) )
        gtk_entry_set_text( ModNameEntry, ModuleXmlInfos->GetName().c_str() );
    if ( !strcmp( gtk_entry_get_text( ModUrlEntry ), "" ) )
        gtk_entry_set_text( ModUrlEntry, ModuleXmlInfos->GetURL().c_str() );
    if ( !strcmp( gtk_entry_get_text( ModXsaEntry ), "" ) )
        gtk_entry_set_text( ModXsaEntry, ModuleXmlInfos->GetXsaURL().c_str() );
    if ( !strcmp( gtk_entry_get_text( ModUsageEntry ), "" ) )
        gtk_entry_set_text( ModUsageEntry,
                            ModuleXmlInfos->GetUsage( "en" ).c_str() );

    textviewSetText( ModAboutTextView,
                     ModuleXmlInfos->GetAbout( "en" ).c_str() );

    // Show short description from modinfo
    if ( ModuleModinfos )
        textviewSetText( ModUsageModinfoTextView,
                         ModuleModinfos->GetUsage( "en" ).c_str() );

    // Populate tree view
    TreeViewMgr->PopulateListStore();

    // Clean-up parameters' entry fields
    gtk_label_set_text( ParamNameLabel, "" );
    gtk_entry_set_text( ParamValuesEntry, "" );
    gtk_entry_set_text( ParamMinEntry, "" );
    gtk_entry_set_text( ParamMaxEntry, "" );
    textviewSetText( ParamModinfoTextView, "" );
    textviewSetText( ParamDescTextView, "" );
}

//--------------------------------------------------------------------------//
//  Start a new module XML file from scratch
//--------------------------------------------------------------------------//
void
ModuleXmlManager::StartNew( const string& filename )
{
    initDruid();
    ModuleXmlFileName = filename;
    ModuleXmlInfos = new Module("");
    TreeViewMgr->SetModXmlData( ModuleXmlInfos );

    if ( Verbosity )
        cout << "Starting module XML druid..." << endl;

    populateDruid();

    // Activate modinfo lookup button
    gtk_widget_set_sensitive( ModinfoButton, TRUE );
}

//--------------------------------------------------------------------------//
//  Load an existing module XML file
//--------------------------------------------------------------------------//
bool
ModuleXmlManager::OpenModuleXml( const string& filename )
{
    initDruid();

    if ( Verbosity )
        cout << "Starting module XML druid with file " << filename
             << "..." << endl;

    ModuleXmlFileName = filename;
    ifstream ifs( filename.c_str() );
    if ( !ifs )
    {
        gnome_appbar_set_status( StatusBar,
                                 _("Could not read module XML file") );
        return false;
    }

    ModuleInfoReader reader( Verbosity );

    // Get infos from XML
    ModuleXmlInfos = reader.ReadFromStream( ifs );
    if ( !ModuleXmlInfos )
    {
        gnome_appbar_set_status( StatusBar,
                                 _("Could not read module XML file") );
        return false;
    }

    gnome_appbar_set_status( StatusBar,
                             _("Successfully read module XML file") );

    // Get infos from modinfo, too
    ModuleModinfos = reader.ReadFromModinfo( ModuleXmlInfos->GetName() );
    if ( !ModuleModinfos )
    {
        gnome_appbar_set_status( StatusBar,
                                 _("Could not get informations from modinfo") );
        return false;
    }

    TreeViewMgr->SetModXmlData( ModuleXmlInfos );
    TreeViewMgr->SetModInfoData( ModuleModinfos );
    populateDruid();

    // Desactivate modinfo lookup button
    gtk_widget_set_sensitive( ModinfoButton, FALSE );

    return true;
}

//--------------------------------------------------------------------------//
//  Write out module XML file
//--------------------------------------------------------------------------//
void
ModuleXmlManager::SaveXmlFile()
{
    // Can't write XML out if module name is missing
    const gchar* modName = gtk_entry_get_text( ModNameEntry );
    if ( !modName || !strcmp(modName, "") )
    {
        gnome_appbar_set_status( StatusBar,
            _("The module name is missing. Can't write out the XML file.") );
        return;
    }

    ModuleXmlInfos->SetName( modName );

    // Read all the data from the druid's fields, and update ModuleXmlInfos
    const gchar* modUrl = gtk_entry_get_text( ModUrlEntry );
    const gchar* modXsaUrl = gtk_entry_get_text( ModXsaEntry );
    const gchar* modUsage = gtk_entry_get_text( ModUsageEntry );

    if ( modUrl )
        ModuleXmlInfos->SetURL( modUrl );
    if ( modXsaUrl )
        ModuleXmlInfos->SetXsaURL( modXsaUrl );
    if ( modUsage )
        ModuleXmlInfos->SetUsage( modUsage, "en" );

    gchar* about = textviewGetText( ModAboutTextView );
    if ( about )
        ModuleXmlInfos->SetAbout( about, "en" );
    g_free( about );

    GtkTreeModel* model = GTK_TREE_MODEL( TreeViewMgr->GetListStore() );
    GtkTreeIter iter;
    if ( gtk_tree_model_get_iter_first( model, &iter ) )
    {
        do
        {
            gchar *name, *type, *values, *defaultVal, *minVal, *maxVal, *desc;
            gtk_tree_model_get( model, &iter,
                                MOD_PARAM_NAME, &name,
                                MOD_PARAM_TYPE, &type,
                                MOD_PARAM_VALUES, &values,
                                MOD_PARAM_DEFAULT, &defaultVal,
                                MOD_PARAM_MIN, &minVal,
                                MOD_PARAM_MAX, &maxVal,
                                MOD_PARAM_DESC, &desc,
                                -1 );

            Param* p = ModuleXmlInfos->Find( name );
            if ( !p )
            {
                // FIXME: add proper support for parameter groups
                p = new Param( name, type );
                ParameterGroup* pg = new ParameterGroup("");
                pg->AddParam( p ); 
                ModuleXmlInfos->AddGroup( pg );
            }
            else
                p->SetType( type );

            if ( values )
                p->SetValues( values );
            if ( defaultVal )
                p->SetDefault( defaultVal );
            if ( minVal )
                p->SetMin( minVal );
            if ( maxVal )
                p->SetMax( maxVal );
            if ( desc )
                p->SetDescription( desc, "en" );
        }
        while( gtk_tree_model_iter_next( model, &iter ) );
    }

    // Write the module XML file
    ofstream ofs( ModuleXmlFileName.c_str() );
    if ( !ofs )
    {
        gnome_appbar_set_status( StatusBar,
                                 _("Could not write out module XML file") );
        return;
    }
    ModuleInfoWriter modwriter;
    modwriter.WriteToStream( ModuleXmlInfos, ofs );
}

//--------------------------------------------------------------------------//
//  Callbacks
//--------------------------------------------------------------------------//
void
ModuleXmlManager::HeaderModinfoButtonHandler()
{
    string modName = gtk_entry_get_text( ModNameEntry );
    if ( modName.empty() )
        return;

    ModuleInfoReader reader( Verbosity );
    Module* mod = reader.ReadFromModinfo( modName );

    if ( !mod )
    {
        gnome_appbar_set_status( StatusBar,
                                 _("Could not get informations from modinfo") );
    }
    else
    {
        if ( Verbosity )
            cout << "Loading modinfo data into druid" << endl;

        delete ModuleModinfos;
        ModuleModinfos = mod;
        TreeViewMgr->SetModInfoData( ModuleModinfos );
        populateDruid();
    }
}

void
ModuleXmlManager::ParamDeleteButtonHandler()
{
    const gchar* name = gtk_entry_get_text( ModNameEntry );
    TreeViewMgr->RemoveParameter( name );
    TreeViewMgr->DeleteSelectedEntry();
}

void
ModuleXmlManager::ParamUpButtonHandler()
{
    TreeViewMgr->MoveSelectedEntryUp();
}

void
ModuleXmlManager::ParamDownButtonHandler()
{
    TreeViewMgr->MoveSelectedEntryDown();
}

void
ModuleXmlManager::ParamAddButtonHandler()
{
    const gchar* name = gtk_entry_get_text( ModNameEntry );
    if ( TreeViewMgr->InsertParameter( name ) )
        TreeViewMgr->AddEntry();
}

void
ModuleXmlManager::ParamApplyButtonHandler()
{
    TreeViewMgr->UpdateSelectedEntry();
}

void
ModuleXmlManager::ParamClearButtonHandler()
{
    // TODO: revert back to modinfo's original data?
}

void
ModuleXmlManager::ParamTypeChanged()
{
    if ( TreeViewMgr )
        TreeViewMgr->CheckType();
}

