//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  xsagui.cpp: GUI utilities for XSA data
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include "xsagui.h"
#include "xsatypes.h"

using namespace gmodcfg;
using namespace std;

static gchar* columnNames[3] = { _("Name"), _("Version"), _("Last release") };

//--------------------------------------------------------------------------//
//  Build the list store for XSA data
//--------------------------------------------------------------------------//
void
XsaTreeViewManager::createListStore()
{
    // Build the liststore
    ListStore = gtk_list_store_new( XSA_NUM_COLUMNS,
                                    G_TYPE_STRING, G_TYPE_STRING,
                                    G_TYPE_STRING, G_TYPE_STRING,
                                    G_TYPE_STRING, G_TYPE_STRING,
                                    G_TYPE_STRING );
    GtkTreeIter iter;
    XsaData::ProductVector::const_iterator pit = Data->BeginProducts();
    XsaData::ProductVector::const_iterator pitend = Data->EndProducts();
    for( ; pit != pitend; ++pit )
    {
        const XsaData::Product* prod = *pit;
        gtk_list_store_append( ListStore, &iter );
        gtk_list_store_set( ListStore, &iter,
                            XSA_PRODUCT_ID, prod->Id.c_str(),
                            XSA_PRODUCT_NAME, prod->Name.c_str(),
                            XSA_PRODUCT_VERSION, prod->Version.c_str(),
                            XSA_PRODUCT_LASTRELEASE, prod->LastRelease.c_str(),
                            XSA_PRODUCT_URL, prod->InfoUrl.c_str(),
                            XSA_PRODUCT_CHANGES, prod->Changes.c_str(),
                            XSA_PRODUCT_FILE, prod->File.c_str(),
                            -1 );

        InsertProduct( prod->Id.c_str() );
    }
}

//--------------------------------------------------------------------------//
//  Create columns in the tree view
//--------------------------------------------------------------------------//
void
XsaTreeViewManager::initTreeView()
{
    // Create only 3 columns in treeview: name, version, last release date
    for( int i=0; i<3; ++i )
    {
        GtkCellRenderer* renderer = gtk_cell_renderer_text_new();
        GtkTreeViewColumn* column =
            gtk_tree_view_column_new_with_attributes( columnNames[i],
                                        renderer, "text", i, NULL );
        gtk_tree_view_append_column( TreeView, column );
    }
}

//--------------------------------------------------------------------------//
//  Move XSA data from one place to another in the list store
//--------------------------------------------------------------------------//
void
XsaTreeViewManager::moveListItem( GtkTreeIter* from, GtkTreeIter* to )
{
    GtkTreeModel* model = GTK_TREE_MODEL( ListStore );
    gchar *id, *name, *version, *lastrel, *url, *changes, *file;
    gtk_tree_model_get( model, from,
                        XSA_PRODUCT_ID, &id,
                        XSA_PRODUCT_NAME, &name,
                        XSA_PRODUCT_VERSION, &version,
                        XSA_PRODUCT_LASTRELEASE, &lastrel,
                        XSA_PRODUCT_URL, &url,
                        XSA_PRODUCT_CHANGES, &changes,
                        XSA_PRODUCT_FILE, &file,
                        -1 );
    gtk_list_store_set( ListStore, to,
                        XSA_PRODUCT_ID, id,
                        XSA_PRODUCT_NAME, name,
                        XSA_PRODUCT_VERSION, version,
                        XSA_PRODUCT_LASTRELEASE, lastrel,
                        XSA_PRODUCT_URL, url,
                        XSA_PRODUCT_CHANGES, changes,
                        XSA_PRODUCT_FILE, file,
                        -1 );
    gtk_list_store_remove( ListStore, from );
}

//--------------------------------------------------------------------------//
//  Time formats conversion functions
//--------------------------------------------------------------------------//
static string
timeToString( const time_t t )
{
    struct tm timePtr;
    gmtime_r( &t, &timePtr );
    char str[16];
    strftime( str, 16, "%Y%m%d", &timePtr );
    return string(str);
}

static time_t
stringToTime( const char* str )
{
    int year, month, day;
    sscanf( str, "%4d%2d%2d", &year, &month, &day );
    struct tm timePtr = { 0, 0, 0, day, month-1, year-1900, 0, 0, -1 };
    return mktime( &timePtr );
}

//--------------------------------------------------------------------------//
//  Update XSA data in the list store
//--------------------------------------------------------------------------//
void
XsaTreeViewManager::setListEntry( GtkTreeIter* iter )
{
    if ( !ProdIdEntry || !ProdNameEntry || !ProdVersionEntry || !ProdUrlEntry
         || !ProdChangesEntry || !ProdFileEntry || !ProdLastRelEntry )
        return;

    // Check that all necessary fields are non-null
    // Note: "file" is not mandatory, as the module might not be packaged yet
    //       the way gmodconfig expects
    const gchar* id = gtk_entry_get_text( ProdIdEntry );
    const gchar* name = gtk_entry_get_text( ProdNameEntry );
    const gchar* version = gtk_entry_get_text( ProdVersionEntry );
    if ( !strcmp(id, "") || !strcmp(name, "") || !strcmp(version, "") )
        return;

    string lastrel = timeToString( gnome_date_edit_get_time(ProdLastRelEntry) );
    const gchar* url = gtk_entry_get_text( ProdUrlEntry );
    const gchar* changes = gtk_entry_get_text( ProdChangesEntry );
    const gchar* file = gtk_entry_get_text( ProdFileEntry );

    gtk_list_store_set( ListStore, iter,
                        XSA_PRODUCT_ID, id,
                        XSA_PRODUCT_NAME, name,
                        XSA_PRODUCT_VERSION, version,
                        XSA_PRODUCT_LASTRELEASE, lastrel.c_str(),
                        XSA_PRODUCT_URL, url,
                        XSA_PRODUCT_CHANGES, changes,
                        XSA_PRODUCT_FILE, file,
                        -1 );
}

//--------------------------------------------------------------------------//
//  Display details of a product
//--------------------------------------------------------------------------//
void
XsaTreeViewManager::ShowProduct( GtkTreeIter* iter )
{
    GtkTreeModel* model = GTK_TREE_MODEL( ListStore );
    gchar *id, *name, *version, *lastrel, *url, *changes, *file;
    gtk_tree_model_get( model, iter,
                        XSA_PRODUCT_ID, &id,
                        XSA_PRODUCT_NAME, &name,
                        XSA_PRODUCT_VERSION, &version,
                        XSA_PRODUCT_LASTRELEASE, &lastrel,
                        XSA_PRODUCT_URL, &url,
                        XSA_PRODUCT_CHANGES, &changes,
                        XSA_PRODUCT_FILE, &file,
                        -1 );
    gtk_entry_set_text( ProdIdEntry, id );
    gtk_entry_set_text( ProdNameEntry, name );
    gtk_entry_set_text( ProdVersionEntry, version );
    gnome_date_edit_set_time( ProdLastRelEntry, stringToTime( lastrel ) );
    gtk_entry_set_text( ProdUrlEntry, url );
    gtk_entry_set_text( ProdChangesEntry, changes );
    gtk_entry_set_text( ProdFileEntry, file );
}

//--------------------------------------------------------------------------//
//  Return filename of selected entry if any, else 0
//--------------------------------------------------------------------------//
const gchar*
XsaTreeViewManager::GetSelectedFile() const
{
    GtkTreeSelection* selection = gtk_tree_view_get_selection( TreeView );
    GtkTreeIter iter;
    if ( gtk_tree_selection_get_selected( selection, NULL, &iter ) )
    {
        GtkTreeModel* model = GTK_TREE_MODEL( ListStore );
        gchar *file;
        gtk_tree_model_get( model, &iter, XSA_PRODUCT_FILE, &file, -1 );
        return file;
    }
    return 0;
}

//--------------------------------------------------------------------------//
//  Insert product id (to avoid duplicates in the treeview)
//  Returns true for success.
//--------------------------------------------------------------------------//
bool
XsaTreeViewManager::InsertProduct( const gchar* prodId )
{
    if ( !prodId || strcmp( prodId, "" ) == 0 )
        return false;
    pair< set<string>::iterator, bool > p = Entries.insert( prodId );
    return p.second;
}

//--------------------------------------------------------------------------//
//  Remove product id (to avoid duplicates in the treeview)
//--------------------------------------------------------------------------//
void
XsaTreeViewManager::RemoveProduct( const gchar* prodId )
{
    Entries.erase( prodId );
}

