//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  xsamgr.cpp: XSA druid
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "xsamgr.h"
#include "xsagui.h"
#include "xsatypes.h"
#include "xsareader.h"
#include "xsawriter.h"
#include "editinterface.h"
#include "support.h"

#include <fstream>
#include <iostream>

using namespace gmodcfg;
using namespace std;

XsaManager::XsaManager( GtkWidget* appWidget, int verbosity )
    : Assistant(0), Druid(0), XsaStuff(0), Verbosity(verbosity),
      TreeView(0), AuthorNameEntry(0), AuthorEmailEntry(0), AuthorUrlEntry(0),
      ProdIdEntry(0), ProdNameEntry(0), ProdVersionEntry(0),
      ProdLastRelEntry(0),ProdUrlEntry(0), ProdChangesEntry(0),
      ProdFileEntry(0)
{
    StatusBar = GNOME_APPBAR( lookup_widget( appWidget, "editappbar" ) );
}

XsaManager::~XsaManager()
{
    delete XsaStuff;
}

//--------------------------------------------------------------------------//
//  Builds the druid
//--------------------------------------------------------------------------//
void
XsaManager::initDruid()
{
    if ( !XsaStuff )
    {
        if ( Verbosity )
            cout << "No XsaData object, don't start the druid" << endl;
        return;
    }

    Assistant = create_xsawindow();
    Druid = GNOME_DRUID( lookup_widget( Assistant, "xsadruid" ) );

    initAuthorPage();
    initProductsPage();
    gtk_widget_show( Assistant );
}

//--------------------------------------------------------------------------//
//  Fill in author infos, if available
//--------------------------------------------------------------------------//
void
XsaManager::initAuthorPage()
{
    AuthorNameEntry =
        GTK_ENTRY( lookup_widget( Assistant, "xsadruid_vendor_name" ) );
    AuthorEmailEntry =
        GTK_ENTRY( lookup_widget( Assistant, "xsadruid_vendor_email" ) );
    AuthorUrlEntry =
        GTK_ENTRY( lookup_widget( Assistant, "xsadruid_vendor_website" ) );

    gtk_entry_set_text( AuthorNameEntry, XsaStuff->GetVendorName().c_str() );
    gtk_entry_set_text( AuthorEmailEntry, XsaStuff->GetVendorEmail().c_str() );
    gtk_entry_set_text( AuthorUrlEntry, XsaStuff->GetVendorUrl().c_str() );
}

//--------------------------------------------------------------------------//
//  Treeview's "changed" callback
//--------------------------------------------------------------------------//
static void
selectionChangedCallback( GtkTreeSelection* select, gpointer data )
{
    GtkTreeView* treeView = gtk_tree_selection_get_tree_view( select );
    GtkTreeModel* model = gtk_tree_view_get_model( treeView );
    XsaTreeViewManager* treeMgr = static_cast<XsaTreeViewManager*>(data);
    GtkTreeIter iter;
    if ( gtk_tree_selection_get_selected( select, &model, &iter ) )
        treeMgr->ShowProduct( &iter );
}

//--------------------------------------------------------------------------//
//  Fill in products infos, if available
//--------------------------------------------------------------------------//
void
XsaManager::initProductsPage()
{
    ProdIdEntry = GTK_ENTRY(
                lookup_widget( Assistant, "xsadruid_products_entry_id" ) );
    ProdNameEntry = GTK_ENTRY(
                lookup_widget( Assistant, "xsadruid_products_entry_name" ) );
    ProdVersionEntry = GTK_ENTRY(
                lookup_widget( Assistant, "xsadruid_products_entry_version" ) );
    ProdLastRelEntry = GNOME_DATE_EDIT(
                lookup_widget( Assistant, "xsadruid_products_entry_date" ) );
    ProdUrlEntry = GTK_ENTRY(
                lookup_widget( Assistant, "xsadruid_products_entry_url" ) );
    ProdChangesEntry = GTK_ENTRY(
                lookup_widget( Assistant, "xsadruid_products_entry_changes" ) );
    ProdFileEntry = GTK_ENTRY(
                lookup_widget( Assistant, "xsadruid_products_entry_file" ) );

    // Set up the treeview manager
    TreeView = GTK_TREE_VIEW( lookup_widget( Assistant,
                                            "xsadruid_products_treeview" ) );
    TreeViewMgr = new XsaTreeViewManager( TreeView, ProdIdEntry, ProdNameEntry,
                    ProdVersionEntry, ProdUrlEntry, ProdChangesEntry,
                    ProdFileEntry, ProdLastRelEntry, XsaStuff );
    TreeViewMgr->InitListStore();

    // Install a callback to detect clicks on the treeview
    GtkTreeSelection* select = gtk_tree_view_get_selection( TreeView );
    g_signal_connect( G_OBJECT( select ), "changed",
                      G_CALLBACK( selectionChangedCallback ),
                      (gpointer)TreeViewMgr );
}

//--------------------------------------------------------------------------//
//  Start a new XSA file from scratch
//--------------------------------------------------------------------------//
void
XsaManager::StartNew( const string& filename )
{
    delete XsaStuff;
    XsaStuff = 0;

    XsaFileName = filename;
    XsaStuff = new XsaData();
    initDruid();
}

//--------------------------------------------------------------------------//
//  Load an existing XSA file
//--------------------------------------------------------------------------//
bool
XsaManager::OpenXsa( const string& filename )
{
    delete XsaStuff;
    XsaStuff = 0;

    if ( Verbosity )
        cout << "Opening " << filename << endl;

    XsaFileName = filename;
    ifstream ifs( filename.c_str() );
    if ( ifs.good() )
    {
        XsaReader reader( Verbosity );
        XsaStuff = reader.Read( ifs );
    }

    if ( !XsaStuff )
    {
        gnome_appbar_set_status( StatusBar, _("Could not read XSA file") );
        return false;
    }

    gnome_appbar_set_status( StatusBar, _("Successfully read XSA file") );
    initDruid();
    return true;
}

//--------------------------------------------------------------------------//
//  Write out XSA file
//--------------------------------------------------------------------------//
void
XsaManager::SaveXsaFile()
{
    // Read all the data from the druid's fields, and update XsaStuff
    const gchar* name = gtk_entry_get_text( AuthorNameEntry );
    const gchar* email = gtk_entry_get_text( AuthorEmailEntry );
    const gchar* url = gtk_entry_get_text( AuthorUrlEntry );

    if ( name )
        XsaStuff->SetVendorName( name );
    if ( email )
        XsaStuff->SetVendorEmail( email );
    if ( url )
        XsaStuff->SetVendorUrl( url );

    XsaStuff->ClearProducts();

    GtkTreeModel* model = GTK_TREE_MODEL( TreeViewMgr->GetListStore() );
    GtkTreeIter iter;
    if ( gtk_tree_model_get_iter_first( model, &iter ) )
    {
        do
        {
            gchar *id, *name, *version, *lastrel, *url, *changes, *file;
            gtk_tree_model_get( model, &iter,
                                XSA_PRODUCT_ID, &id,
                                XSA_PRODUCT_NAME, &name,
                                XSA_PRODUCT_VERSION, &version,
                                XSA_PRODUCT_LASTRELEASE, &lastrel,
                                XSA_PRODUCT_URL, &url,
                                XSA_PRODUCT_CHANGES, &changes,
                                XSA_PRODUCT_FILE, &file,
                                -1 );
            if ( Verbosity )
            {
                cout << "Id=" << id << "\nName=" << name << "\nVersion="
                     << version << "\nLastRel=" << lastrel << "\nUrl="
                     << url << "\nChanges=" << changes << endl;
            }
            XsaStuff->AddProduct( id, name, version, lastrel, url, changes,
                                  file );
        }
        while( gtk_tree_model_iter_next( model, &iter ) );
    }

    // Write the XSA file
    ofstream ofs( XsaFileName.c_str() );
    if ( !ofs )
    {
        // TODO: emit message to status bar
        cout << _("Could not write out XSA file") << endl;
        return;
    }
    XsaWriter writer( Verbosity );
    writer.Write( XsaStuff, ofs );
}

//--------------------------------------------------------------------------//
//  Callbacks
//--------------------------------------------------------------------------//
void
XsaManager::ProductsDeleteButtonHandler()
{
    const gchar* id = gtk_entry_get_text( ProdIdEntry );
    TreeViewMgr->RemoveProduct( id );
    TreeViewMgr->DeleteSelectedEntry();
}

void
XsaManager::ProductsUpButtonHandler()
{
    TreeViewMgr->MoveSelectedEntryUp();
}

void
XsaManager::ProductsDownButtonHandler()
{
    TreeViewMgr->MoveSelectedEntryDown();
}

void
XsaManager::ProductsAddButtonHandler()
{
    const gchar* id = gtk_entry_get_text( ProdIdEntry );
    if ( TreeViewMgr->InsertProduct( id ) )
        TreeViewMgr->AddEntry();
}

void
XsaManager::ProductsApplyButtonHandler()
{
    GtkTreeModel* model = gtk_tree_view_get_model( TreeView );
    GtkTreeSelection* selection = gtk_tree_view_get_selection( TreeView );
    GtkTreeIter iter;
    if ( gtk_tree_selection_get_selected( selection, NULL, &iter ) )
    {
        // Compare old id (in liststore) to new one (in entry widget)
        gchar* oldid;
        gtk_tree_model_get( model, &iter, XSA_PRODUCT_ID, &oldid, -1 );
        const gchar* newid = gtk_entry_get_text( ProdIdEntry );

        if ( strcmp( oldid, newid ) == 0 )
        {
            TreeViewMgr->UpdateSelectedEntry();
        }
        else if ( TreeViewMgr->InsertProduct( newid ) )
        {
            TreeViewMgr->RemoveProduct( oldid );
            TreeViewMgr->UpdateSelectedEntry();
        }
    }
}

void
XsaManager::ProductsClearButtonHandler()
{
    gtk_entry_set_text( ProdIdEntry, "" );
    gtk_entry_set_text( ProdNameEntry, "" );
    gtk_entry_set_text( ProdVersionEntry, "" );
    gtk_entry_set_text( ProdUrlEntry, "" );
    gtk_entry_set_text( ProdChangesEntry, "" );
    gtk_entry_set_text( ProdFileEntry, "" );
}

