//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  xsareader.cpp: read an XSA file from an istream
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "xsareader.h"
#include "xsatypes.h"
#include <gnome.h>
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <iterator>
#include <sstream>
#include <libxml/parser.h>
#include "xmlutils.h"

using std::cout;
using std::endl;
using std::istream;

using namespace gmodcfg;

//--------------------------------------------------------------------------//
//  Helper function to parse a vendor definition in an XSA's XML tree
//--------------------------------------------------------------------------//
static void
parseVendor( XsaData* xsaData, xmlDoc* doc, xmlNode* node )
{
    // Parse the children nodes.
    for( xmlNode* iter = node->xmlChildrenNode; iter; iter = iter->next )
    {
        string text = XmlGetContent( doc, iter );
        if ( text.empty() )
            continue;

        if ( !xmlStrcmp( iter->name, (const xmlChar*)"name" ) )
            xsaData->SetVendorName( text );
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"email" ) )
            xsaData->SetVendorEmail( text );
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"url" ) )
            xsaData->SetVendorUrl( text );
    }
}

//--------------------------------------------------------------------------//
//  Helper function to parse a product definition in an XSA's XML tree
//  Returns: ptr to allocated Product object
//--------------------------------------------------------------------------//
static void
parseProduct( XsaData* xsaData, xmlDoc* doc, xmlNode* node, int verbosity )
{
    string id = XmlGetAttribute( node, "id" );
    string name, version, lastrel, url, changes, file;

    // Parse the child nodes.
    for( xmlNode* iter = node->xmlChildrenNode; iter; iter = iter->next )
    {
        string text = XmlGetContent( doc, iter );
        if ( text.empty() )
            continue;

        if ( !xmlStrcmp( iter->name, (const xmlChar*)"name" ) )
            name = text;
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"version" ) )
            version = text;
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"last-release" ) )
            lastrel = text;
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"info-url" ) )
            url = text;
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"changes" ) )
            changes = text;
        else if ( !xmlStrcmp( iter->name, (const xmlChar*)"file" ) )
            file = text;
    }

    xsaData->AddProduct( id, name, version, lastrel, url, changes, file );

    if ( verbosity >= 3 )
        cout << "product: name=" << name
             << "\n         version=" << version
             << "\n         file=" << file << endl;
}

//--------------------------------------------------------------------------//
//  Read in XSA from a stream
//  Returns: ptr to an allocated XsaData object
//--------------------------------------------------------------------------//
static XsaData*
readXsaFromStream( istream& is, int verbosity )
{
    xmlDoc* doc = XmlParseFromStream( is );
    xmlNode* node = xmlDocGetRootElement( doc );
    if ( !node )
        throw std::runtime_error("Empty XML document");

    if ( verbosity )
        cout << "Reading XSA,,," << endl;

    // Traverse the document tree. First level must be <xsa>
    if ( xmlStrcmp( node->name, (const xmlChar *)"xsa" ) )
        throw std::runtime_error("Missing top level node \"xsa\"");

    XsaData* xsaData = new XsaData;

    // 2nd level is either vendor, or product
    node = node->xmlChildrenNode;
    if ( !node )
        throw std::runtime_error("No XSA data available");

    for( ; node; node = node->next )
    {
        if ( !xmlStrcmp( node->name, (const xmlChar*)"vendor" ) )
        {
            parseVendor( xsaData, doc, node );
            if ( verbosity >= 3 )
                cout << "vendor: name=" << xsaData->GetVendorName()
                     << "\n        email=" << xsaData->GetVendorEmail()
                     << "\n        url=" << xsaData->GetVendorUrl() << endl;
        }
        else if ( !xmlStrcmp( node->name, (const xmlChar*)"product" ) )
            parseProduct( xsaData, doc, node, verbosity );
    }

    xmlFreeDoc( doc );
    return xsaData;
}

//--------------------------------------------------------------------------//
//  Main entry point
//  Returns: a new XsaData object if success, else NULL
//--------------------------------------------------------------------------//
XsaData*
XsaReader::Read( istream& is ) const
{
    XsaData* xsaData = NULL;

    try
    {
        xsaData = readXsaFromStream( is, Verbosity );
    }
    catch( const std::exception& e )
    {
        if ( Verbosity )
            cout << "Exception caught in XsaReader: " << e.what() << endl;
    }

    if ( Verbosity )
        cout << "Done reading XSA" << endl;

    return xsaData;
}

